import os
import time
import ctypes

from threading import Lock

from .remoteApiBindings import vrep as remote_api
from ..robot.io import AbstractIO


vrep_error = {
    remote_api.simx_return_ok: 'Ok',
    remote_api.simx_return_novalue_flag: 'No value',
    remote_api.simx_return_timeout_flag: 'Timeout',
    remote_api.simx_return_illegal_opmode_flag: 'Opmode error',
    remote_api.simx_return_remote_error_flag: 'Remote error',
    remote_api.simx_return_split_progress_flag: 'Progress error',
    remote_api.simx_return_local_error_flag: 'Local error',
    remote_api.simx_return_initialize_error_flag: 'Init error'
}

vrep_mode = {
    'normal': remote_api.simx_opmode_oneshot_wait,
    'streaming': remote_api.simx_opmode_streaming,
    'sending': remote_api.simx_opmode_oneshot,
}


class VrepIO(AbstractIO):

    """ This class is used to get/set values from/to a V-REP scene.

        It is based on V-REP remote API (http://www.coppeliarobotics.com/helpFiles/en/remoteApiOverview.htm).

    """
    MAX_ITER = 5
    TIMEOUT = 0.4

    def __init__(self, vrep_host='127.0.0.1', vrep_port=19997, scene=None, start=False):
        """ Starts the connection with the V-REP remote API server.

        :param str vrep_host: V-REP remote API server host
        :param int vrep_port: V-REP remote API server port
        :param str scene: path to a V-REP scene file
        :param bool start: whether to start the scene after loading it

        .. warning:: Only one connection can be established with the V-REP remote server API. So before trying to connect make sure that all previously started connections have been closed (see :func:`~pypot.vrep.io.close_all_connections`)

        """
        self._object_handles = {}
        self._lock = Lock()

        self.vrep_host = vrep_host
        self.vrep_port = vrep_port
        self.scene = scene
        self.start = start

        # self.client_id = remote_api.simxStart(
        #     vrep_host, vrep_port, True, True, 5000, 5)
        # if self.client_id == -1:
        #     msg = ('Could not connect to V-REP server on {}:{}. '
        #            'This could also means that you still have '
        #            'a previously opened connection running! '
        #            '(try pypot.vrep.close_all_connections())')
        #     raise VrepConnectionError(msg.format(vrep_host, vrep_port))

        # if scene is not None:
        #     self.load_scene(scene, start)

        self.open_io()

    def open_io(self):
        self.client_id = remote_api.simxStart(
            self.vrep_host, self.vrep_port, True, True, 5000, 5)
        if self.client_id == -1:
            msg = ('Could not connect to V-REP server on {}:{}. '
                   'This could also means that you still have '
                   'a previously opened connection running! '
                   '(try pypot.vrep.close_all_connections())')
            raise VrepConnectionError(
                msg.format(self.vrep_host, self.vrep_port))

        if self.scene is not None:
            self.load_scene(self.scene, self.start)

    def close(self):
        """ Closes the current connection. """
        with self._lock:
            remote_api.simxFinish(self.client_id)

    def load_scene(self, scene_path, start=False):
        """ Loads a scene on the V-REP server.

        :param str scene_path: path to a V-REP scene file
        :param bool start: whether to directly start the simulation after loading the scene

        .. note:: It is assumed that the scene file is always available on the server side.

        """
        self.stop_simulation()

        if not os.path.exists(scene_path):
            raise IOError("No such file or directory: '{}'".format(scene_path))

        self.call_remote_api('simxLoadScene', scene_path, True)

        if start:
            self.start_simulation()

    def start_simulation(self):
        """ Starts the simulation.

            .. note:: Do nothing if the simulation is already started.

            .. warning:: if you start the simulation just after stopping it, the simulation will likely not be started. Use :meth:`~pypot.vrep.io.VrepIO.restart_simulation` instead.
        """
        self.call_remote_api('simxStartSimulation')

        # We have to force a sleep
        # Otherwise it may causes troubles??
        time.sleep(0.5)

    def restart_simulation(self):
        """ Re-starts the simulation. """
        self.stop_simulation()
        # We have to force a sleep
        # Otherwise the simulation is not restarted
        time.sleep(0.5)
        self.start_simulation()

    def stop_simulation(self):
        """ Stops the simulation. """
        self.call_remote_api('simxStopSimulation')

    def pause_simulation(self):
        """ Pauses the simulation. """
        self.call_remote_api('simxPauseSimulation')

    def resume_simulation(self):
        """ Resumes the simulation. """
        self.start_simulation()

    def get_motor_position(self, motor_name):
        """ Gets the motor current position. """
        return self.call_remote_api('simxGetJointPosition',
                                    self.get_object_handle(motor_name),
                                    streaming=True)

    def set_motor_position(self, motor_name, position):
        """ Sets the motor target position. """
        self.call_remote_api('simxSetJointTargetPosition',
                             self.get_object_handle(motor_name),
                             position,
                             sending=True)

    def get_motor_force(self, motor_name):
        """ Retrieves the force or torque applied to a joint along/about its active axis. """
        return self.call_remote_api('simxGetJointForce',
                                    self.get_object_handle(motor_name),
                                    streaming=True)

    def set_motor_force(self, motor_name, force):
        """  Sets the maximum force or torque that a joint can exert. """
        self.call_remote_api('simxSetJointForce',
                             self.get_object_handle(motor_name),
                             force,
                             sending=True)

    def get_object_position(self, object_name, relative_to_object=None):
        """ Gets the object position. """
        h = self.get_object_handle(object_name)
        relative_handle = (-1 if relative_to_object is None
                           else self.get_object_handle(relative_to_object))

        return self.call_remote_api('simxGetObjectPosition',
                                    h, relative_handle,
                                    streaming=True)

    def set_object_position(self, object_name, position=[0, 0, 0]):
        """ Sets the object position. """
        h = self.get_object_handle(object_name)

        return self.call_remote_api('simxSetObjectPosition',
                                    h, -1, position,
                                    sending=True)

    def get_object_orientation(self, object_name, relative_to_object=None):
        """ Gets the object orientation. """
        h = self.get_object_handle(object_name)
        relative_handle = (-1 if relative_to_object is None
                           else self.get_object_handle(relative_to_object))

        return self.call_remote_api('simxGetObjectOrientation',
                                    h, relative_handle,
                                    streaming=True)

    def _get_object_handle(self, obj):
        return self.call_remote_api('simxGetObjectHandle', obj)

    def get_object_handle(self, obj):
        """ Gets the vrep object handle. """
        if obj not in self._object_handles:
            self._object_handles[obj] = self._get_object_handle(obj=obj)

        return self._object_handles[obj]

    def get_collision_state(self, collision_name):
        """ Gets the collision state. """
        return self.call_remote_api('simxReadCollision',
                                    self.get_collision_handle(collision_name),
                                    streaming=True)

    def _get_collision_handle(self, collision):
        return self.call_remote_api('simxGetCollisionHandle', collision)

    def get_collision_handle(self, collision):
        """ Gets a vrep collisions handle. """
        if collision not in self._object_handles:
            h = self._get_collision_handle(collision)
            self._object_handles[collision] = h

        return self._object_handles[collision]

    def get_simulation_current_time(self, timer='CurrentTime'):
        """ Gets the simulation current time. """
        try:
            return self.call_remote_api('simxGetFloatSignal', timer, streaming=True)
        except VrepIOErrors:
            return 0.0

    def add_cube(self, name, position, sizes, mass):
        """ Add Cube """
        self._create_pure_shape(0, 239, sizes, mass, [0, 0])
        self.set_object_position("Cuboid", position)
        self.change_object_name("Cuboid", name)

    def add_sphere(self, name, position, sizes, mass, precision=[10, 10]):
        """ Add Sphere """
        self._create_pure_shape(1, 239, sizes, mass, precision)
        self.set_object_position("Sphere", position)
        self.change_object_name("Sphere", name)

    def add_cylinder(self, name, position, sizes, mass, precision=[10, 10]):
        """ Add Cylinder """
        self._create_pure_shape(2, 239, sizes, mass, precision)
        self.set_object_position("Cylinder", position)
        self.change_object_name("Cylinder", name)

    def add_cone(self, name, position, sizes, mass, precision=[10, 10]):
        """ Add Cone """
        self._create_pure_shape(3, 239, sizes, mass, precision)
        self.set_object_position("Cylinder", position)
        self.change_object_name("Cylinder", name)

    def change_object_name(self, old_name, new_name):
        """ Change object name """
        h = self._get_object_handle(old_name)
        if old_name in self._object_handles:
            self._object_handles.pop(old_name)
        lua_code = "simSetObjectName({}, '{}')".format(h, new_name)
        self._inject_lua_code(lua_code)

    def _create_pure_shape(self, primitive_type, options, sizes, mass, precision):
        """ Create Pure Shape """
        lua_code = "simCreatePureShape({}, {}, {{{}, {}, {}}}, {}, {{{}, {}}})".format(
            primitive_type, options, sizes[0], sizes[1], sizes[2], mass, precision[0], precision[1])
        self._inject_lua_code(lua_code)

    def _inject_lua_code(self, lua_code):
        """ Sends raw lua code and evaluate it without any checking! """
        msg = (ctypes.c_ubyte * len(lua_code)).from_buffer_copy(lua_code.encode())
        self.call_remote_api('simxWriteStringStream', 'my_lua_code', msg)

    def call_remote_api(self, func_name, *args, **kwargs):
        """ Calls any remote API func in a thread_safe way.

        :param str func_name: name of the remote API func to call
        :param args: args to pass to the remote API call
        :param kwargs: args to pass to the remote API call

        .. note:: You can add an extra keyword to specify if you want to use the streaming or sending mode. The oneshot_wait mode is used by default (see `here <http://www.coppeliarobotics.com/helpFiles/en/remoteApiConstants.htm#operationModes>`_ for details about possible modes).

        .. warning:: You should not pass the clientId and the operationMode as arguments. They will be automatically added.

        As an example you can retrieve all joints name using the following call::

            vrep_io.remote_api_call('simxGetObjectGroupData',
                                    vrep_io.remote_api.sim_object_joint_type,
                                    0,
                                    streaming=True)

        """
        f = getattr(remote_api, func_name)

        mode = self._extract_mode(kwargs)
        kwargs['operationMode'] = vrep_mode[mode]
        # hard_retry = True

        if '_force' in kwargs:
            del kwargs['_force']
            _force = True
        else:
            _force = False

        for _ in range(VrepIO.MAX_ITER):
            with self._lock:
                ret = f(self.client_id, *args, **kwargs)

            if _force:
                return

            if mode == 'sending' or isinstance(ret, int):
                err, res = ret, None
            else:
                err, res = ret[0], ret[1:]
                res = res[0] if len(res) == 1 else res

            err = [bool((err >> i) & 1) for i in range(len(vrep_error))]

            if remote_api.simx_return_novalue_flag not in err:
                break

            time.sleep(VrepIO.TIMEOUT)

        # if any(err) and hard_retry:
        #     print "HARD RETRY"
        # self.stop_simulation() #nope
        #
        #     notconnected = True
        #     while notconnected:
        #         self.close()
        #         close_all_connections()
        #         time.sleep(0.5)
        #         try:
        #             self.open_io()
        #             notconnected = False
        #         except:
        #             print 'CONNECTION ERROR'
        #             pass
        #
        #     self.start_simulation()
        #
        #     with self._lock:
        #         ret = f(self.client_id, *args, **kwargs)
        #
        #         if mode == 'sending' or isinstance(ret, int):
        #             err, res = ret, None
        #         else:
        #             err, res = ret[0], ret[1:]
        #             res = res[0] if len(res) == 1 else res
        #
        #         err = [bool((err >> i) & 1) for i in range(len(vrep_error))]
        #
        #         return res

        if any(err):
            msg = ' '.join([vrep_error[2 ** i]
                            for i, e in enumerate(err) if e])
            raise VrepIOErrors(msg)

        return res

    def _extract_mode(self, kwargs):
        for mode in ('streaming', 'sending'):
            if mode in kwargs:
                kwargs.pop(mode)
                return mode

        return 'normal'


def close_all_connections():
    """ Closes all opened connection to V-REP remote API server. """
    remote_api.simxFinish(-1)


# V-REP Errors
class VrepIOError(Exception):

    """ Base class for V-REP IO Errors. """

    def __init__(self, error_code, message):
        message = 'V-REP error code {} ({}): "{}"'.format(
            error_code, vrep_error[error_code], message)
        Exception.__init__(self, message)


class VrepIOErrors(Exception):
    pass


class VrepConnectionError(Exception):

    """ Base class for V-REP connection Errors. """
    pass



# This file is part of the REMOTE API
# 
# Copyright 2006-2016 Coppelia Robotics GmbH. All rights reserved. 
# marc@coppeliarobotics.com
# www.coppeliarobotics.com
# 
# The REMOTE API is licensed under the terms of GNU GPL:
# 
# -------------------------------------------------------------------
# The REMOTE API is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# THE REMOTE API IS DISTRIBUTED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTY. THE USER WILL USE IT AT HIS/HER OWN RISK. THE ORIGINAL
# AUTHORS AND COPPELIA ROBOTICS GMBH WILL NOT BE LIABLE FOR DATA LOSS,
# DAMAGES, LOSS OF PROFITS OR ANY OTHER KIND OF LOSS WHILE USING OR
# MISUSING THIS SOFTWARE.
# 
# See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with the REMOTE API.  If not, see <http://www.gnu.org/licenses/>.
# -------------------------------------------------------------------
#
# This file was automatically created for V-REP release V3.3.2 on August 29th 2016

#constants
#Scene object types. Values are serialized
sim_object_shape_type           =0
sim_object_joint_type           =1
sim_object_graph_type           =2
sim_object_camera_type          =3
sim_object_dummy_type           =4
sim_object_proximitysensor_type =5
sim_object_reserved1            =6
sim_object_reserved2            =7
sim_object_path_type            =8
sim_object_visionsensor_type    =9
sim_object_volume_type          =10
sim_object_mill_type            =11
sim_object_forcesensor_type     =12
sim_object_light_type           =13
sim_object_mirror_type          =14

#General object types. Values are serialized
sim_appobj_object_type          =109
sim_appobj_collision_type       =110
sim_appobj_distance_type        =111
sim_appobj_simulation_type      =112
sim_appobj_ik_type              =113
sim_appobj_constraintsolver_type=114
sim_appobj_collection_type      =115
sim_appobj_ui_type              =116
sim_appobj_script_type          =117
sim_appobj_pathplanning_type    =118
sim_appobj_RESERVED_type        =119
sim_appobj_texture_type         =120

# Ik calculation methods. Values are serialized
sim_ik_pseudo_inverse_method        =0
sim_ik_damped_least_squares_method  =1
sim_ik_jacobian_transpose_method    =2

# Ik constraints. Values are serialized
sim_ik_x_constraint         =1
sim_ik_y_constraint         =2
sim_ik_z_constraint         =4
sim_ik_alpha_beta_constraint=8
sim_ik_gamma_constraint     =16
sim_ik_avoidance_constraint =64

# Ik calculation results 
sim_ikresult_not_performed  =0
sim_ikresult_success        =1
sim_ikresult_fail           =2

# Scene object sub-types. Values are serialized 
# Light sub-types 
sim_light_omnidirectional_subtype   =1
sim_light_spot_subtype              =2
sim_light_directional_subtype       =3
# Joint sub-types 
sim_joint_revolute_subtype          =10
sim_joint_prismatic_subtype         =11
sim_joint_spherical_subtype         =12
# Shape sub-types 
sim_shape_simpleshape_subtype       =20
sim_shape_multishape_subtype        =21
# Proximity sensor sub-types 
sim_proximitysensor_pyramid_subtype =30
sim_proximitysensor_cylinder_subtype=31
sim_proximitysensor_disc_subtype    =32
sim_proximitysensor_cone_subtype    =33
sim_proximitysensor_ray_subtype     =34
# Mill sub-types 
sim_mill_pyramid_subtype            =40
sim_mill_cylinder_subtype           =41
sim_mill_disc_subtype               =42
sim_mill_cone_subtype               =42
# No sub-type 
sim_object_no_subtype               =200


#Scene object main properties (serialized)
sim_objectspecialproperty_collidable                    =0x0001
sim_objectspecialproperty_measurable                    =0x0002
#reserved                        =0x0004 
#reserved                        =0x0008 
sim_objectspecialproperty_detectable_ultrasonic            =0x0010
sim_objectspecialproperty_detectable_infrared            =0x0020
sim_objectspecialproperty_detectable_laser                =0x0040
sim_objectspecialproperty_detectable_inductive            =0x0080
sim_objectspecialproperty_detectable_capacitive            =0x0100
sim_objectspecialproperty_renderable                    =0x0200
sim_objectspecialproperty_detectable_all =sim_objectspecialproperty_detectable_ultrasonic|sim_objectspecialproperty_detectable_infrared|sim_objectspecialproperty_detectable_laser|sim_objectspecialproperty_detectable_inductive|sim_objectspecialproperty_detectable_capacitive
sim_objectspecialproperty_cuttable                        =0x0400
sim_objectspecialproperty_pathplanning_ignored            =0x0800

# Model properties (serialized)
sim_modelproperty_not_collidable                =0x0001
sim_modelproperty_not_measurable                =0x0002
sim_modelproperty_not_renderable                =0x0004
sim_modelproperty_not_detectable                =0x0008
sim_modelproperty_not_cuttable                    =0x0010
sim_modelproperty_not_dynamic                    =0x0020
sim_modelproperty_not_respondable                =0x0040 # cannot be selected if sim_modelproperty_not_dynamic is not selected 
sim_modelproperty_not_reset                        =0x0080 # Model is not reset at simulation end. This flag is cleared at simulation end 
sim_modelproperty_not_visible                    =0x0100 # Whole model is invisible independent of local visibility settings 
sim_modelproperty_not_model                        =0xf000 # object is not a model 


# Check the documentation instead of comments below!! 
# Following messages are dispatched to the Lua-message container 
sim_message_ui_button_state_change  =0    # a UI button slider etc. changed (due to a user's action). aux[0]=UI handle aux[1]=button handle aux[2]=button attributes aux[3]=slider position (if slider) 
sim_message_reserved9               =1    # Do not use 
sim_message_object_selection_changed=2
sim_message_reserved10                =3    # do not use 
sim_message_model_loaded            =4
sim_message_reserved11                =5    # do not use 
sim_message_keypress                =6    # a key was pressed while the focus was on a page (aux[0]=key aux[1]=ctrl and shift key state) 
sim_message_bannerclicked            =7    # a banner was clicked (aux[0]=banner ID) 


# Following messages are dispatched only to the C-API (not available from Lua) 
sim_message_for_c_api_only_start        =0x100      # Do not use 
sim_message_reserved1                   =0x101      # Do not use 
sim_message_reserved2                    =0x102      # Do not use 
sim_message_reserved3                    =0x103      # Do not use 
sim_message_eventcallback_scenesave        =0x104        # about to save a scene 
sim_message_eventcallback_modelsave        =0x105      # about to save a model (current selection will be saved) 
sim_message_eventcallback_moduleopen    =0x106        # called when simOpenModule in Lua is called 
sim_message_eventcallback_modulehandle    =0x107        # called when simHandleModule in Lua is called with argument false 
sim_message_eventcallback_moduleclose    =0x108        # called when simCloseModule in Lua is called 
sim_message_reserved4                    =0x109      # Do not use 
sim_message_reserved5                    =0x10a        # Do not use 
sim_message_reserved6                    =0x10b        # Do not use 
sim_message_reserved7                    =0x10c        # Do not use 
sim_message_eventcallback_instancepass    =0x10d        # Called once every main application loop pass. auxiliaryData[0] contains event flags of events that happened since last time 
sim_message_eventcallback_broadcast     =0x10e
sim_message_eventcallback_imagefilter_enumreset =0x10f
sim_message_eventcallback_imagefilter_enumerate      =0x110
sim_message_eventcallback_imagefilter_adjustparams   =0x111
sim_message_eventcallback_imagefilter_reserved       =0x112
sim_message_eventcallback_imagefilter_process        =0x113
sim_message_eventcallback_reserved1                  =0x114   # do not use 
sim_message_eventcallback_reserved2                  =0x115   # do not use 
sim_message_eventcallback_reserved3                  =0x116   # do not use 
sim_message_eventcallback_reserved4                  =0x117   # do not use 
sim_message_eventcallback_abouttoundo                 =0x118   # the undo button was hit and a previous state is about to be restored 
sim_message_eventcallback_undoperformed                 =0x119   # the undo button was hit and a previous state restored 
sim_message_eventcallback_abouttoredo                 =0x11a   # the redo button was hit and a future state is about to be restored  
sim_message_eventcallback_redoperformed                 =0x11b   # the redo button was hit and a future state restored  
sim_message_eventcallback_scripticondblclick         =0x11c   # script icon was double clicked.  (aux[0]=object handle associated with script set replyData[0] to 1 if script should not be opened)  
sim_message_eventcallback_simulationabouttostart     =0x11d
sim_message_eventcallback_simulationended            =0x11e
sim_message_eventcallback_reserved5                     =0x11f   # do not use 
sim_message_eventcallback_keypress                     =0x120   # a key was pressed while the focus was on a page (aux[0]=key aux[1]=ctrl and shift key state) 
sim_message_eventcallback_modulehandleinsensingpart  =0x121   # called when simHandleModule in Lua is called with argument true 
sim_message_eventcallback_renderingpass              =0x122   # called just before the scene is rendered 
sim_message_eventcallback_bannerclicked              =0x123   # called when a banner was clicked (aux[0]=banner ID) 
sim_message_eventcallback_menuitemselected           =0x124   # auxiliaryData[0] indicates the handle of the item auxiliaryData[1] indicates the state of the item 
sim_message_eventcallback_refreshdialogs             =0x125   # aux[0]=refresh degree (0=light 1=medium 2=full) 
sim_message_eventcallback_sceneloaded                =0x126
sim_message_eventcallback_modelloaded                =0x127
sim_message_eventcallback_instanceswitch             =0x128
sim_message_eventcallback_guipass                    =0x129
sim_message_eventcallback_mainscriptabouttobecalled  =0x12a
sim_message_eventcallback_rmlposition                =0x12b   #the command simRMLPosition was called. The appropriate plugin should handle the call
sim_message_eventcallback_rmlvelocity                =0x12c   # the command simRMLVelocity was called. The appropriate plugin should handle the call
sim_message_simulation_start_resume_request          =0x1000
sim_message_simulation_pause_request                 =0x1001
sim_message_simulation_stop_request                  =0x1002

# Scene object properties. Combine with the | operator 
sim_objectproperty_reserved1                =0x0000
sim_objectproperty_reserved2                =0x0001
sim_objectproperty_reserved3                =0x0002
sim_objectproperty_reserved4                =0x0003
sim_objectproperty_reserved5                =0x0004 # formerly sim_objectproperty_visible 
sim_objectproperty_reserved6                =0x0008 # formerly sim_objectproperty_wireframe 
sim_objectproperty_collapsed                =0x0010
sim_objectproperty_selectable                =0x0020
sim_objectproperty_reserved7                =0x0040
sim_objectproperty_selectmodelbaseinstead    =0x0080
sim_objectproperty_dontshowasinsidemodel    =0x0100
# reserved                                    =0x0200 
sim_objectproperty_canupdatedna                =0x0400
sim_objectproperty_selectinvisible            =0x0800
sim_objectproperty_depthinvisible            =0x1000


# type of arguments (input and output) for custom lua commands 
sim_lua_arg_nil     =0
sim_lua_arg_bool    =1    
sim_lua_arg_int     =2
sim_lua_arg_float   =3
sim_lua_arg_string  =4
sim_lua_arg_invalid =5
sim_lua_arg_table   =8

# custom user interface properties. Values are serialized. 
sim_ui_property_visible                        =0x0001
sim_ui_property_visibleduringsimulationonly    =0x0002
sim_ui_property_moveable                    =0x0004
sim_ui_property_relativetoleftborder        =0x0008
sim_ui_property_relativetotopborder            =0x0010
sim_ui_property_fixedwidthfont                =0x0020
sim_ui_property_systemblock                    =0x0040
sim_ui_property_settocenter                    =0x0080
sim_ui_property_rolledup                    =0x0100
sim_ui_property_selectassociatedobject        =0x0200
sim_ui_property_visiblewhenobjectselected    =0x0400


# button properties. Values are serialized. 
sim_buttonproperty_button                =0x0000
sim_buttonproperty_label                =0x0001
sim_buttonproperty_slider                =0x0002
sim_buttonproperty_editbox                =0x0003
sim_buttonproperty_staydown                =0x0008
sim_buttonproperty_enabled                =0x0010
sim_buttonproperty_borderless            =0x0020
sim_buttonproperty_horizontallycentered    =0x0040
sim_buttonproperty_ignoremouse            =0x0080
sim_buttonproperty_isdown                =0x0100
sim_buttonproperty_transparent            =0x0200
sim_buttonproperty_nobackgroundcolor    =0x0400
sim_buttonproperty_rollupaction            =0x0800
sim_buttonproperty_closeaction            =0x1000
sim_buttonproperty_verticallycentered    =0x2000
sim_buttonproperty_downupevent            =0x4000


# Simulation status 
sim_simulation_stopped                        =0x00                                # Simulation is stopped 
sim_simulation_paused                        =0x08                                # Simulation is paused 
sim_simulation_advancing                    =0x10                                # Simulation is advancing 
sim_simulation_advancing_firstafterstop        =sim_simulation_advancing|0x00        # First simulation pass (1x) 
sim_simulation_advancing_running            =sim_simulation_advancing|0x01        # Normal simulation pass (>=1x) 
# reserved                                    =sim_simulation_advancing|0x02 
sim_simulation_advancing_lastbeforepause    =sim_simulation_advancing|0x03        # Last simulation pass before pause (1x) 
sim_simulation_advancing_firstafterpause    =sim_simulation_advancing|0x04        # First simulation pass after pause (1x) 
sim_simulation_advancing_abouttostop        =sim_simulation_advancing|0x05        # "Trying to stop" simulation pass (>=1x) 
sim_simulation_advancing_lastbeforestop        =sim_simulation_advancing|0x06        # Last simulation pass (1x) 


# Script execution result (first return value) 
sim_script_no_error                    =0
sim_script_main_script_nonexistent    =1
sim_script_main_script_not_called    =2
sim_script_reentrance_error            =4
sim_script_lua_error                =8
sim_script_call_error                =16


 # Script types (serialized!) 
sim_scripttype_mainscript   =0
sim_scripttype_childscript  =1
sim_scripttype_jointctrlcallback  =4
sim_scripttype_contactcallback  =5
sim_scripttype_customizationscript  =6
sim_scripttype_generalcallback  =7

# API call error messages 
sim_api_errormessage_ignore    =0    # does not memorize nor output errors 
sim_api_errormessage_report    =1    # memorizes errors (default for C-API calls) 
sim_api_errormessage_output    =2  # memorizes and outputs errors (default for Lua-API calls) 


# special argument of some functions 
sim_handle_all                        =-2
sim_handle_all_except_explicit        =-3
sim_handle_self                        =-4
sim_handle_main_script                =-5
sim_handle_tree                        =-6
sim_handle_chain                    =-7
sim_handle_single                    =-8
sim_handle_default                    =-9
sim_handle_all_except_self            =-10
sim_handle_parent                    =-11


# special handle flags
sim_handleflag_assembly                =0x400000
sim_handleflag_model                =0x800000


# distance calculation methods (serialized) 
sim_distcalcmethod_dl               =0
sim_distcalcmethod_dac              =1
sim_distcalcmethod_max_dl_dac       =2
sim_distcalcmethod_dl_and_dac       =3
sim_distcalcmethod_sqrt_dl2_and_dac2=4
sim_distcalcmethod_dl_if_nonzero    =5
sim_distcalcmethod_dac_if_nonzero   =6


 # Generic dialog styles 
sim_dlgstyle_message        =0
sim_dlgstyle_input          =1
sim_dlgstyle_ok             =2
sim_dlgstyle_ok_cancel      =3
sim_dlgstyle_yes_no         =4
sim_dlgstyle_dont_center    =32# can be combined with one of above values. Only with this flag can the position of the related UI be set just after dialog creation  

 # Generic dialog return values 
sim_dlgret_still_open   =0
sim_dlgret_ok           =1
sim_dlgret_cancel       =2
sim_dlgret_yes          =3
sim_dlgret_no           =4


# Path properties 
sim_pathproperty_show_line                            =0x0001
sim_pathproperty_show_orientation                    =0x0002
sim_pathproperty_closed_path                        =0x0004
sim_pathproperty_automatic_orientation                =0x0008
sim_pathproperty_invert_velocity                    =0x0010
sim_pathproperty_infinite_acceleration                =0x0020
sim_pathproperty_flat_path                            =0x0040
sim_pathproperty_show_position                        =0x0080
sim_pathproperty_auto_velocity_profile_translation    =0x0100
sim_pathproperty_auto_velocity_profile_rotation        =0x0200
sim_pathproperty_endpoints_at_zero                    =0x0400
sim_pathproperty_keep_x_up                            =0x0800


 # drawing objects 
# following are mutually exclusive 
sim_drawing_points          =0            # 3 values per point (point size in pixels) 
sim_drawing_lines            =1            # 6 values per line (line size in pixels) 
sim_drawing_triangles        =2            # 9 values per triangle 
sim_drawing_trianglepoints    =3            # 6 values per point (3 for triangle position 3 for triangle normal vector) (triangle size in meters) 
sim_drawing_quadpoints        =4            # 6 values per point (3 for quad position 3 for quad normal vector) (quad size in meters) 
sim_drawing_discpoints        =5            # 6 values per point (3 for disc position 3 for disc normal vector) (disc size in meters) 
sim_drawing_cubepoints        =6          # 6 values per point (3 for cube position 3 for cube normal vector) (cube size in meters) 
sim_drawing_spherepoints    =7          # 3 values per point (sphere size in meters) 

# following can be or-combined 
sim_drawing_itemcolors                =0x00020 # +3 values per item (each item has its own ambient color (rgb values)).
                                             # Mutually exclusive with sim_drawing_vertexcolors 
sim_drawing_vertexcolors            =0x00040 # +3 values per vertex (each vertex has its own ambient color (rgb values). Only for sim_drawing_lines (+6) and for sim_drawing_triangles(+9)). Mutually exclusive with sim_drawing_itemcolors 
sim_drawing_itemsizes                =0x00080 # +1 value per item (each item has its own size). Not for sim_drawing_triangles 
sim_drawing_backfaceculling            =0x00100 # back faces are not displayed for all items 
sim_drawing_wireframe                =0x00200 # all items displayed in wireframe 
sim_drawing_painttag                =0x00400 # all items are tagged as paint (for additinal processing at a later stage) 
sim_drawing_followparentvisibility    =0x00800 # if the object is associated with a scene object then it follows that visibility otherwise it is always visible 
sim_drawing_cyclic                    =0x01000 # if the max item count was reached then the first items are overwritten. 
sim_drawing_50percenttransparency    =0x02000 # the drawing object will be 50% transparent 
sim_drawing_25percenttransparency    =0x04000 # the drawing object will be 25% transparent 
sim_drawing_12percenttransparency    =0x08000 # the drawing object will be 12.5% transparent 
sim_drawing_emissioncolor            =0x10000 # When used in combination with sim_drawing_itemcolors or sim_drawing_vertexcolors then the specified colors will be for the emissive component 
sim_drawing_facingcamera            =0x20000 # Only for trianglepoints quadpoints discpoints and cubepoints. If specified the normal verctor is calculated to face the camera (each item data requires 3 values less) 
sim_drawing_overlay                    =0x40000 # When specified objects are always drawn on top of "regular objects" 
sim_drawing_itemtransparency        =0x80000  # +1 value per item (each item has its own transparency value (0-1)). Not compatible with sim_drawing_vertexcolors 

# banner values 
# following can be or-combined 
sim_banner_left                        =0x00001 # Banners display on the left of the specified point 
sim_banner_right                    =0x00002 # Banners display on the right of the specified point 
sim_banner_nobackground                =0x00004 # Banners have no background rectangle 
sim_banner_overlay                    =0x00008 # When specified banners are always drawn on top of "regular objects" 
sim_banner_followparentvisibility    =0x00010 # if the object is associated with a scene object then it follows that visibility otherwise it is always visible 
sim_banner_clickselectsparent        =0x00020 # if the object is associated with a scene object then clicking the banner will select the scene object 
sim_banner_clicktriggersevent        =0x00040 # if the banner is clicked an event is triggered (sim_message_eventcallback_bannerclicked and sim_message_bannerclicked are generated) 
sim_banner_facingcamera                =0x00080 # If specified the banner will always face the camera by rotating around the banner's vertical axis (y-axis) 
sim_banner_fullyfacingcamera        =0x00100 # If specified the banner will always fully face the camera (the banner's orientation is same as the camera looking at it) 
sim_banner_backfaceculling            =0x00200 # If specified the banner will only be visible from one side 
sim_banner_keepsamesize                =0x00400 # If specified the banner will always appear in the same size. In that case size represents the character height in pixels 
sim_banner_bitmapfont                =0x00800 # If specified a fixed-size bitmap font is used. The text will also always fully face the camera and be right 
                                             # to the specified position. Bitmap fonts are not clickable 


# particle objects following are mutually exclusive 
sim_particle_points1        =0  # 6 values per point (pt1 and pt2. Pt1 is start position pt2-pt1 is the initial velocity vector). i
                                #Point is 1 pixel big. Only appearance is a point internally handled as a perfect sphere 
sim_particle_points2        =1    # 6 values per point. Point is 2 pixel big. Only appearance is a point internally handled as a perfect sphere 
sim_particle_points4        =2    # 6 values per point. Point is 4 pixel big. Only appearance is a point internally handled as a perfect sphere 
sim_particle_roughspheres    =3    # 6 values per sphere. Only appearance is rough. Internally a perfect sphere 
sim_particle_spheres        =4    # 6 values per sphere. Internally a perfect sphere 




# following can be or-combined 
sim_particle_respondable1to4        =0x0020 # the particles are respondable against shapes (against all objects that have at least one bit 1-4 activated in the global respondable mask) 
sim_particle_respondable5to8        =0x0040 # the particles are respondable against shapes (against all objects that have at least one bit 5-8 activated in the global respondable mask) 
sim_particle_particlerespondable    =0x0080 # the particles are respondable against each other 
sim_particle_ignoresgravity            =0x0100 # the particles ignore the effect of gravity. Not compatible with sim_particle_water 
sim_particle_invisible                =0x0200 # the particles are invisible 
sim_particle_itemsizes                =0x0400 # +1 value per particle (each particle can have a different size) 
sim_particle_itemdensities            =0x0800 # +1 value per particle (each particle can have a different density) 
sim_particle_itemcolors                =0x1000 # +3 values per particle (each particle can have a different color) 
sim_particle_cyclic                    =0x2000 # if the max item count was reached then the first items are overwritten. 
sim_particle_emissioncolor            =0x4000 # When used in combination with sim_particle_itemcolors then the specified colors will be for the emissive component 
sim_particle_water                    =0x8000 # the particles are water particles (no weight in the water (i.e. when z<0)). Not compatible with sim_particle_ignoresgravity 
sim_particle_painttag                =0x10000 # The particles can be seen by vision sensors (sim_particle_invisible must not be set) 




# custom user interface menu attributes 
sim_ui_menu_title        =1
sim_ui_menu_minimize    =2
sim_ui_menu_close        =4
sim_ui_menu_systemblock =8



# Boolean parameters 
sim_boolparam_hierarchy_visible                 =0
sim_boolparam_console_visible                   =1
sim_boolparam_collision_handling_enabled        =2
sim_boolparam_distance_handling_enabled         =3
sim_boolparam_ik_handling_enabled               =4
sim_boolparam_gcs_handling_enabled              =5
sim_boolparam_dynamics_handling_enabled         =6
sim_boolparam_joint_motion_handling_enabled     =7
sim_boolparam_path_motion_handling_enabled      =8
sim_boolparam_proximity_sensor_handling_enabled =9
sim_boolparam_vision_sensor_handling_enabled    =10
sim_boolparam_mill_handling_enabled             =11
sim_boolparam_browser_visible                   =12
sim_boolparam_scene_and_model_load_messages     =13
sim_reserved0                                   =14
sim_boolparam_shape_textures_are_visible        =15
sim_boolparam_display_enabled                   =16
sim_boolparam_infotext_visible                  =17
sim_boolparam_statustext_open                   =18
sim_boolparam_fog_enabled                       =19
sim_boolparam_rml2_available                    =20
sim_boolparam_rml4_available                    =21
sim_boolparam_mirrors_enabled                    =22
sim_boolparam_aux_clip_planes_enabled            =23
sim_boolparam_full_model_copy_from_api            =24
sim_boolparam_realtime_simulation                =25
sim_boolparam_force_show_wireless_emission        =27
sim_boolparam_force_show_wireless_reception        =28
sim_boolparam_video_recording_triggered            =29
sim_boolparam_threaded_rendering_enabled        =32
sim_boolparam_fullscreen                        =33
sim_boolparam_headless                            =34
sim_boolparam_hierarchy_toolbarbutton_enabled    =35
sim_boolparam_browser_toolbarbutton_enabled        =36
sim_boolparam_objectshift_toolbarbutton_enabled    =37
sim_boolparam_objectrotate_toolbarbutton_enabled=38
sim_boolparam_force_calcstruct_all_visible        =39
sim_boolparam_force_calcstruct_all                =40
sim_boolparam_exit_request                        =41
sim_boolparam_play_toolbarbutton_enabled        =42
sim_boolparam_pause_toolbarbutton_enabled        =43
sim_boolparam_stop_toolbarbutton_enabled        =44
sim_boolparam_waiting_for_trigger                =45


# Integer parameters 
sim_intparam_error_report_mode      =0  # Check sim_api_errormessage_... constants above for valid values 
sim_intparam_program_version        =1  # e.g Version 2.1.4 --> 20104. Can only be read 
sim_intparam_instance_count         =2  # do not use anymore (always returns 1 since V-REP 2.5.11) 
sim_intparam_custom_cmd_start_id    =3  # can only be read 
sim_intparam_compilation_version    =4  # 0=evaluation version 1=full version 2=player version. Can only be read 
sim_intparam_current_page           =5
sim_intparam_flymode_camera_handle  =6  # can only be read 
sim_intparam_dynamic_step_divider   =7  # can only be read 
sim_intparam_dynamic_engine         =8  # 0=Bullet 1=ODE. 2=Vortex.
sim_intparam_server_port_start      =9  # can only be read 
sim_intparam_server_port_range      =10 # can only be read 
sim_intparam_visible_layers         =11
sim_intparam_infotext_style         =12
sim_intparam_settings               =13
sim_intparam_edit_mode_type         =14 # can only be read 
sim_intparam_server_port_next       =15 # is initialized at sim_intparam_server_port_start 
sim_intparam_qt_version             =16 # version of the used Qt framework 
sim_intparam_event_flags_read       =17 # can only be read 
sim_intparam_event_flags_read_clear =18 # can only be read 
sim_intparam_platform               =19 # can only be read 
sim_intparam_scene_unique_id        =20 # can only be read 
sim_intparam_work_thread_count      =21
sim_intparam_mouse_x                =22
sim_intparam_mouse_y                =23
sim_intparam_core_count             =24
sim_intparam_work_thread_calc_time_ms =25
sim_intparam_idle_fps               =26
sim_intparam_prox_sensor_select_down =27
sim_intparam_prox_sensor_select_up  =28
sim_intparam_stop_request_counter   =29
sim_intparam_program_revision       =30
sim_intparam_mouse_buttons          =31
sim_intparam_dynamic_warning_disabled_mask =32
sim_intparam_simulation_warning_disabled_mask =33
sim_intparam_scene_index            =34
sim_intparam_motionplanning_seed    =35
sim_intparam_speedmodifier          =36

# Float parameters 
sim_floatparam_rand=0 # random value (0.0-1.0) 
sim_floatparam_simulation_time_step =1
sim_floatparam_stereo_distance        =2

# String parameters 
sim_stringparam_application_path=0 # path of V-REP's executable 
sim_stringparam_video_filename=1
sim_stringparam_app_arg1            =2
sim_stringparam_app_arg2            =3
sim_stringparam_app_arg3            =4
sim_stringparam_app_arg4            =5
sim_stringparam_app_arg5            =6
sim_stringparam_app_arg6            =7
sim_stringparam_app_arg7            =8
sim_stringparam_app_arg8            =9
sim_stringparam_app_arg9            =10
sim_stringparam_scene_path_and_name    =13

# Array parameters 
sim_arrayparam_gravity          =0
sim_arrayparam_fog              =1
sim_arrayparam_fog_color        =2
sim_arrayparam_background_color1=3
sim_arrayparam_background_color2=4
sim_arrayparam_ambient_light    =5
sim_arrayparam_random_euler        =6

sim_objintparam_visibility_layer= 10
sim_objfloatparam_abs_x_velocity= 11
sim_objfloatparam_abs_y_velocity= 12
sim_objfloatparam_abs_z_velocity= 13
sim_objfloatparam_abs_rot_velocity= 14
sim_objfloatparam_objbbox_min_x= 15
sim_objfloatparam_objbbox_min_y= 16
sim_objfloatparam_objbbox_min_z= 17
sim_objfloatparam_objbbox_max_x= 18
sim_objfloatparam_objbbox_max_y= 19
sim_objfloatparam_objbbox_max_z= 20
sim_objfloatparam_modelbbox_min_x= 21
sim_objfloatparam_modelbbox_min_y= 22
sim_objfloatparam_modelbbox_min_z= 23
sim_objfloatparam_modelbbox_max_x= 24
sim_objfloatparam_modelbbox_max_y= 25
sim_objfloatparam_modelbbox_max_z= 26
sim_objintparam_collection_self_collision_indicator= 27
sim_objfloatparam_transparency_offset= 28
sim_objintparam_child_role= 29
sim_objintparam_parent_role= 30
sim_objintparam_manipulation_permissions= 31
sim_objintparam_illumination_handle= 32

sim_visionfloatparam_near_clipping= 1000
sim_visionfloatparam_far_clipping= 1001
sim_visionintparam_resolution_x= 1002
sim_visionintparam_resolution_y= 1003
sim_visionfloatparam_perspective_angle= 1004
sim_visionfloatparam_ortho_size= 1005
sim_visionintparam_disabled_light_components= 1006
sim_visionintparam_rendering_attributes= 1007
sim_visionintparam_entity_to_render= 1008
sim_visionintparam_windowed_size_x= 1009
sim_visionintparam_windowed_size_y= 1010
sim_visionintparam_windowed_pos_x= 1011
sim_visionintparam_windowed_pos_y= 1012
sim_visionintparam_pov_focal_blur= 1013
sim_visionfloatparam_pov_blur_distance= 1014
sim_visionfloatparam_pov_aperture= 1015
sim_visionintparam_pov_blur_sampled= 1016
sim_visionintparam_render_mode= 1017

sim_jointintparam_motor_enabled= 2000
sim_jointintparam_ctrl_enabled= 2001
sim_jointfloatparam_pid_p= 2002
sim_jointfloatparam_pid_i= 2003
sim_jointfloatparam_pid_d= 2004
sim_jointfloatparam_intrinsic_x= 2005
sim_jointfloatparam_intrinsic_y= 2006
sim_jointfloatparam_intrinsic_z= 2007
sim_jointfloatparam_intrinsic_qx= 2008
sim_jointfloatparam_intrinsic_qy= 2009
sim_jointfloatparam_intrinsic_qz= 2010
sim_jointfloatparam_intrinsic_qw= 2011
sim_jointfloatparam_velocity= 2012
sim_jointfloatparam_spherical_qx= 2013
sim_jointfloatparam_spherical_qy= 2014
sim_jointfloatparam_spherical_qz= 2015
sim_jointfloatparam_spherical_qw= 2016
sim_jointfloatparam_upper_limit= 2017
sim_jointfloatparam_kc_k= 2018
sim_jointfloatparam_kc_c= 2019
sim_jointfloatparam_ik_weight= 2021
sim_jointfloatparam_error_x= 2022
sim_jointfloatparam_error_y= 2023
sim_jointfloatparam_error_z= 2024
sim_jointfloatparam_error_a= 2025
sim_jointfloatparam_error_b= 2026
sim_jointfloatparam_error_g= 2027
sim_jointfloatparam_error_pos= 2028
sim_jointfloatparam_error_angle= 2029
sim_jointintparam_velocity_lock= 2030
sim_jointintparam_vortex_dep_handle= 2031
sim_jointfloatparam_vortex_dep_multiplication= 2032
sim_jointfloatparam_vortex_dep_offset= 2033

sim_shapefloatparam_init_velocity_x= 3000
sim_shapefloatparam_init_velocity_y= 3001
sim_shapefloatparam_init_velocity_z= 3002
sim_shapeintparam_static= 3003
sim_shapeintparam_respondable= 3004
sim_shapefloatparam_mass= 3005
sim_shapefloatparam_texture_x= 3006
sim_shapefloatparam_texture_y= 3007
sim_shapefloatparam_texture_z= 3008
sim_shapefloatparam_texture_a= 3009
sim_shapefloatparam_texture_b= 3010
sim_shapefloatparam_texture_g= 3011
sim_shapefloatparam_texture_scaling_x= 3012
sim_shapefloatparam_texture_scaling_y= 3013
sim_shapeintparam_culling= 3014
sim_shapeintparam_wireframe= 3015
sim_shapeintparam_compound= 3016
sim_shapeintparam_convex= 3017
sim_shapeintparam_convex_check= 3018
sim_shapeintparam_respondable_mask= 3019
sim_shapefloatparam_init_velocity_a= 3020
sim_shapefloatparam_init_velocity_b= 3021
sim_shapefloatparam_init_velocity_g= 3022
sim_shapestringparam_color_name= 3023
sim_shapeintparam_edge_visibility= 3024
sim_shapefloatparam_shading_angle= 3025
sim_shapefloatparam_edge_angle= 3026
sim_shapeintparam_edge_borders_hidden= 3027

sim_proxintparam_ray_invisibility= 4000

sim_forcefloatparam_error_x= 5000
sim_forcefloatparam_error_y= 5001
sim_forcefloatparam_error_z= 5002
sim_forcefloatparam_error_a= 5003
sim_forcefloatparam_error_b= 5004
sim_forcefloatparam_error_g= 5005
sim_forcefloatparam_error_pos= 5006
sim_forcefloatparam_error_angle= 5007

sim_lightintparam_pov_casts_shadows= 8000

sim_cameraintparam_disabled_light_components= 9000
sim_camerafloatparam_perspective_angle= 9001
sim_camerafloatparam_ortho_size= 9002
sim_cameraintparam_rendering_attributes= 9003
sim_cameraintparam_pov_focal_blur= 9004
sim_camerafloatparam_pov_blur_distance= 9005
sim_camerafloatparam_pov_aperture= 9006
sim_cameraintparam_pov_blur_samples= 9007

sim_dummyintparam_link_type= 10000

sim_mirrorfloatparam_width= 12000
sim_mirrorfloatparam_height= 12001
sim_mirrorfloatparam_reflectance= 12002
sim_mirrorintparam_enable= 12003

sim_pplanfloatparam_x_min= 20000
sim_pplanfloatparam_x_range= 20001
sim_pplanfloatparam_y_min= 20002
sim_pplanfloatparam_y_range= 20003
sim_pplanfloatparam_z_min= 20004
sim_pplanfloatparam_z_range= 20005
sim_pplanfloatparam_delta_min= 20006
sim_pplanfloatparam_delta_range= 20007

sim_mplanintparam_nodes_computed= 25000
sim_mplanintparam_prepare_nodes= 25001
sim_mplanintparam_clear_nodes= 25002

# User interface elements 
sim_gui_menubar                        =0x0001
sim_gui_popups                        =0x0002
sim_gui_toolbar1                    =0x0004
sim_gui_toolbar2                    =0x0008
sim_gui_hierarchy                    =0x0010
sim_gui_infobar                        =0x0020
sim_gui_statusbar                    =0x0040
sim_gui_scripteditor                =0x0080
sim_gui_scriptsimulationparameters    =0x0100
sim_gui_dialogs                        =0x0200
sim_gui_browser                        =0x0400
sim_gui_all                            =0xffff


# Joint modes 
sim_jointmode_passive       =0
sim_jointmode_motion        =1
sim_jointmode_ik            =2
sim_jointmode_ikdependent   =3
sim_jointmode_dependent     =4
sim_jointmode_force         =5


# Navigation and selection modes with the mouse. Lower byte values are mutually exclusive upper byte bits can be combined 
sim_navigation_passive                    =0x0000
sim_navigation_camerashift                =0x0001
sim_navigation_camerarotate                =0x0002
sim_navigation_camerazoom                =0x0003
sim_navigation_cameratilt                =0x0004
sim_navigation_cameraangle                =0x0005
sim_navigation_camerafly                =0x0006
sim_navigation_objectshift                =0x0007
sim_navigation_objectrotate                =0x0008
sim_navigation_reserved2                =0x0009
sim_navigation_reserved3                =0x000A
sim_navigation_jointpathtest            =0x000B
sim_navigation_ikmanip                    =0x000C
sim_navigation_objectmultipleselection    =0x000D
# Bit-combine following values and add them to one of above's values for a valid navigation mode 
sim_navigation_reserved4                =0x0100
sim_navigation_clickselection            =0x0200
sim_navigation_ctrlselection            =0x0400
sim_navigation_shiftselection            =0x0800
sim_navigation_camerazoomwheel            =0x1000
sim_navigation_camerarotaterightbutton    =0x2000



#Remote API constants
SIMX_VERSION                    =0 
# Remote API message header structure 
SIMX_HEADER_SIZE                =18
simx_headeroffset_crc           =0    # 1 simxUShort. Generated by the client or server. The CRC for the message 
simx_headeroffset_version       =2    # 1 byte. Generated by the client or server. The version of the remote API software 
simx_headeroffset_message_id    =3    # 1 simxInt. Generated by the client (and used in a reply by the server) 
simx_headeroffset_client_time   =7    # 1 simxInt. Client time stamp generated by the client (and sent back by the server) 
simx_headeroffset_server_time   =11    # 1 simxInt. Generated by the server when a reply is generated. The server timestamp 
simx_headeroffset_scene_id      =15    # 1 simxUShort. Generated by the server. A unique ID identifying the scene currently displayed 
simx_headeroffset_server_state  =17    # 1 byte. Generated by the server. Bit coded 0 set --> simulation not stopped 1 set --> simulation paused 2 set --> real-time switch on 3-5 edit mode type (0=no edit mode 1=triangle 2=vertex 3=edge 4=path 5=UI)  

# Remote API command header 
SIMX_SUBHEADER_SIZE                 =26
simx_cmdheaderoffset_mem_size       =0    # 1 simxInt. Generated by the client or server. The buffer size of the command. 
simx_cmdheaderoffset_full_mem_size  =4    # 1 simxInt. Generated by the client or server. The full buffer size of the command (applies to split chunks). 
simx_cmdheaderoffset_pdata_offset0  =8    # 1 simxUShort. Generated by the client or server. The amount of data that is part of the command identification. 
simx_cmdheaderoffset_pdata_offset1  =10    # 1 simxInt. Generated by the client or server. The amount of shift of the pure data buffer (applies to split chunks). 
simx_cmdheaderoffset_cmd=14    # 1 simxInt. Generated by the client (and used in a reply by the server). The command combined with the operation mode of the command. 
simx_cmdheaderoffset_delay_or_split =18    # 1 simxUShort. Generated by the client or server. The amount of delay in ms of a continuous command or the max. pure data size to send at once (applies to split commands). 
simx_cmdheaderoffset_sim_time       =20    # 1 simxInt. Generated by the server. The simulation time (in ms) when the command was executed (or 0 if simulation is not running) 
simx_cmdheaderoffset_status         =24    # 1 byte. Generated by the server. (1 bit 0 is set --> error in function execution on server side). The client writes bit 1 if command cannot be overwritten
simx_cmdheaderoffset_reserved       =25    # 1 byte. Not yet used 





# Regular operation modes 
simx_opmode_oneshot                =0x000000 # sends command as one chunk. Reply will also come as one chunk. Doesn't wait for the reply. 
simx_opmode_blocking            =0x010000 # sends command as one chunk. Reply will also come as one chunk. Waits for the reply (_REPLY_WAIT_TIMEOUT_IN_MS is the timeout). 
simx_opmode_oneshot_wait        =0x010000 # sends command as one chunk. Reply will also come as one chunk. Waits for the reply (_REPLY_WAIT_TIMEOUT_IN_MS is the timeout). 
simx_opmode_continuous            =0x020000  
simx_opmode_streaming            =0x020000 # sends command as one chunk. Command will be stored on the server and always executed 
                                                      #(every x ms (as far as possible) where x can be 0-65535. just add x to opmode_continuous). 
                                                      # A reply will be sent continuously each time as one chunk. Doesn't wait for the reply. 

# Operation modes for heavy data 
simx_opmode_oneshot_split        =0x030000   # sends command as several chunks (max chunk size is x bytes where x can be _MIN_SPLIT_AMOUNT_IN_BYTES-65535. Just add x to opmode_oneshot_split). Reply will also come as several chunks. Doesn't wait for the reply.      
simx_opmode_continuous_split    =0x040000
simx_opmode_streaming_split    =0x040000    # sends command as several chunks (max chunk size is x bytes where x can be _MIN_SPLIT_AMOUNT_IN_BYTES-65535. Just add x to opmode_continuous_split). Command will be stored on the server and always executed. A reply will be sent continuously each time as several chunks. Doesn't wait for the reply. 

# Special operation modes 
simx_opmode_discontinue            =0x050000    # removes and cancels all commands stored on the client or server side (also continuous commands) 
simx_opmode_buffer                =0x060000    # doesn't send anything but checks if a reply for the given command is available in the input buffer (i.e. previously received from the server) 
simx_opmode_remove                =0x070000    # doesn't send anything and doesn't return any specific value. It just erases a similar command reply in the inbox (to free some memory) 


# Command return codes 
simx_return_ok                    =0x000000
simx_return_novalue_flag        =0x000001        # input buffer doesn't contain the specified command 
simx_return_timeout_flag        =0x000002        # command reply not received in time for opmode_oneshot_wait operation mode 
simx_return_illegal_opmode_flag    =0x000004        # command doesn't support the specified operation mode 
simx_return_remote_error_flag    =0x000008        # command caused an error on the server side 
simx_return_split_progress_flag    =0x000010        # previous similar command not yet fully processed (applies to opmode_oneshot_split operation modes) 
simx_return_local_error_flag    =0x000020        # command caused an error on the client side 
simx_return_initialize_error_flag    =0x000040        # simxStart was not yet called 

# Following for backward compatibility (same as above) 
simx_error_noerror                =0x000000
simx_error_novalue_flag            =0x000001        # input buffer doesn't contain the specified command 
simx_error_timeout_flag            =0x000002        # command reply not received in time for opmode_oneshot_wait operation mode 
simx_error_illegal_opmode_flag    =0x000004        # command doesn't support the specified operation mode 
simx_error_remote_error_flag    =0x000008        # command caused an error on the server side 
simx_error_split_progress_flag    =0x000010        # previous similar command not yet fully processed (applies to opmode_oneshot_split operation modes) 
simx_error_local_error_flag        =0x000020        # command caused an error on the client side 
simx_error_initialize_error_flag        =0x000040        # simxStart was not yet called 






import sys


def flushed_print(*args, **kwargs):
    """
    Use to replace print(*args, flush=True) that doesn't exist for python<3.3
    """
    print(*args, **kwargs)
    file = kwargs.get('file', sys.stdout)
    file.flush() if file is not None else sys.stdout.flush()



from ikpy.chain import Chain
from ikpy.utils.geometry import rpy_matrix
from ikpy.urdf.URDF import get_chain_from_joints
from numpy import deg2rad, rad2deg, array, arctan2, sqrt


class IKChain(Chain):
    """ Motors chain used for forward and inverse kinematics.
    This class is based on the IK Chain as defined in the IKPY library (https://github.com/Phylliade/ikpy). It
    provides convenient methods to directly create such a chain directly from a Poppy Creature.
    """

    @classmethod
    def from_poppy_creature(cls, poppy, motors, passiv, tip, reversed_motors=[]):
        """ Creates an kinematic chain from motors of a Poppy Creature.
        :param poppy: PoppyCreature used
        :param list motors: list of all motors that composed the kinematic chain
        :param list passiv: list of motors which are passiv in the chain (they will not move)
        :param list tip: [x, y, z] translation of the tip of the chain (in meters)
        :param list reversed_motors: list of motors that should be manually reversed (due to a problem in the URDF?)
        """
        chain_elements = get_chain_from_joints(poppy.urdf_file, [m.name for m in motors])

        activ = [False] + [m not in passiv for m in motors] + [True]

        chain = cls.from_urdf_file(poppy.urdf_file,
                                   base_elements=chain_elements,
                                   last_link_vector=tip,
                                   active_links_mask=activ)

        chain.motors = [getattr(poppy, l.name) for l in chain.links[1:-1]]

        for m, l in zip(chain.motors, chain.links[1:-1]):
            # Force an access to angle limit to retrieve real values
            # This is quite an ugly fix and should be handled better
            m.angle_limit

            bounds = m.__dict__['lower_limit'], m.__dict__['upper_limit']
            l.bounds = tuple(map(rad2deg, bounds))

        chain._reversed = array([(-1 if m in reversed_motors else 1) for m in motors])

        return chain

    @property
    def joints_position(self):
        """ Returns the joints position of all motors in the chain (in degrees). """
        return [m.present_position for m in self.motors]

    # Transformation matrix M:
    # [[ Rx.x, Ry.x, Rz.x, T.x ],      R = M[:3][:3] is the rotation matrix.
    #  [ Rx.y, Ry.y, Rz.y, T.y ],
    #  [ Rx.z, Ry.z, Rz.z, T.z ],      T = M[:3][3] is the translation matrix. It corresponds
    #  [ 0   , 0   , 0,    1  ]]       to the absolute coordinates of the effector

    @property
    def position(self):
        """ Returns the cartesian position of the end of the chain (in meters). """
        angles = self.convert_to_ik_angles(self.joints_position)
        return self.forward_kinematics(angles)[:3, 3]

    @property
    def orientation(self):
        """ Returns the rotation matrix along X axis (values from -1 to 1). """
        angles = self.convert_to_ik_angles(self.joints_position)
        return self.forward_kinematics(angles)[:3, 0]

    @property
    def pose(self):
        """
        Gives the 4x4 afﬁne transformation matrix of the current position
        *Used for debug*
        :return: 4x4 afﬁne transformation matrix (float)
        """
        angles = self.convert_to_ik_angles(self.joints_position)
        return self.forward_kinematics(angles)

    @property
    def rpy(self):
        """
        Gives the rpy values of the current position
        :return: roll, pitch, yaw (float)
        """
        angles = self.convert_to_ik_angles(self.joints_position)
        R = self.forward_kinematics(angles)
        yaw = arctan2(R[2][1], R[1][1])
        pitch = arctan2(-R[3][1], sqrt(R[3][2] ^ 2 + R[3][3] ^ 2))
        roll = arctan2(R[3][2], R[3][3])
        return roll, pitch, yaw

    def rpy_to_rotation_matrix(self, r, p, y):
        """
        converts rpy to a 3x3 rotation matrix
        :param r: roll (float)
        :param p: pitch (float)
        :param y: yaw (float)
        :return: 3x3 rotation matrix
        """
        return rpy_matrix(r, p, y)

    def goto(self, position, orientation, duration, wait=False, accurate=False):
        """ Goes to a given cartesian position.
        :param list position: [x, y, z] representing the target position (in meters)
        :param list orientation: [Rx.x, Rx.y, Rx.z] transformation along X axis (values from -1 to 1)
        :param float duration: move duration
        :param bool wait: whether to wait for the end of the move
        :param bool accurate: trade-off between accurate solution and computation time. By default, use the not so
        accurate but fast version.
        """
        # if len(position) != 3:
        #     raise ValueError('Position should be a list [x, y, z]!')
        self._goto(position, orientation, duration, wait, accurate)

    def _goto(self, position, orientation, duration, wait, accurate):
        """ Goes to a given cartesian pose.
        :param matrix position: [x, y, z] representing the target position (in meters)
        :param list orientation: [Rx.x, Rx.y, Rx.z] transformation along X axis (values from -1 to 1)
        :param float duration: move duration
        :param bool wait: whether to wait for the end of the move
        :param bool accurate: trade-off between accurate solution and computation time. By default, use the not so
        accurate but fast version.
        """
        kwargs = {}
        if not accurate:
            kwargs['max_iter'] = 3

        if orientation is not None:
            shape = array(orientation).shape
            if shape == (3, 3):
                orientation_mode = "all"
            elif shape == (3,):
                orientation_mode = "X"
            else:
                orientation_mode = None
        else:
            orientation_mode = None

        # q0 = self.convert_to_ik_angles(self.joints_position)
        q = self.inverse_kinematics(target_position=position,
                                    target_orientation=orientation,
                                    orientation_mode=orientation_mode,
                                    **kwargs)

        joints = self.convert_from_ik_angles(q)

        last = self.motors[-1]
        for m, pos in list(zip(self.motors, joints)):
            m.goto_position(pos, duration, wait=False if m != last else wait)

    def convert_to_ik_angles(self, joints):
        """ Convert from poppy representation to IKPY internal representation. """
        if len(joints) != len(self.motors):
            raise ValueError('Incompatible data, len(joints) should be {}!'.format(len(self.motors)))

        raw_joints = [(j + m.offset) * (1 if m.direct else -1) for j, m in zip(joints, self.motors)]

        raw_joints *= self._reversed

        return [0] + [deg2rad(j) for j in raw_joints] + [0]

    def convert_from_ik_angles(self, joints):
        """ Convert from IKPY internal representation to poppy representation. """
        if len(joints) != len(self.motors) + 2:
            raise ValueError('Incompatible data, len(joints) should be {}!'.format(len(self.motors) + 2))

        joints = [rad2deg(j) for j in joints[1:-1]]
        joints *= self._reversed

        return [(j * (1 if m.direct else -1)) - m.offset
                for j, m in zip(joints, self.motors)]



#!/usr/bin/env python


import os
import sys
import time
import random
import logging
import argparse
import webbrowser

from contextlib import closing
from argparse import RawTextHelpFormatter

from pypot.server.snap import find_local_ip
from pypot.creatures import installed_poppy_creatures
from pypot.utils import flushed_print as print

from multiprocessing import Process
from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler


def start_poppy_with_services(args):
	params = poppy_params_from_args(args)

	for i in range(5):
		try:
			print('Attempt {} to start the robot...'.format(i + 1))
			return installed_poppy_creatures[args.creature](**params)

		except Exception as e:
			# In case of failure,
			# Give the robot some time to statup, reboot...
			time.sleep(random.random())
			print(e)
	else:
		print('Could not start up the robot...')
		sys.exit(1)


def poppy_params_from_args(args):
	params = {
		'use_snap': args.snap,
		'snap_port': args.snap_port,
		'use_http': args.http or args.scratch,
		'http_port': args.http_port,
		'use_scratch': args.scratch,
		'scratch_port': args.scratch_port,
		'use_remote': args.remote,
		'use_ws': args.ws,
		'ws_port': args.ws_port,
	}

	if args.verbose:
		params['snap_quiet'] = False
		params['http_quiet'] = False
		params['ws_quiet'] = False

	if args.vrep:
		params['simulator'] = 'vrep'
	elif args.poppy_simu:
		params['simulator'] = 'poppy-simu'
	elif args.dummy:
		params['simulator'] = 'dummy'

	if args.disable_camera:
		params['camera'] = 'dummy'

	return params


def main():
	parser = argparse.ArgumentParser(
		description=('Poppy services launcher. Use it to quickly instantiate a poppy creature with Snap!, '
					 'Scratch, an http server, or a remote robot.'),
		epilog="""
Examples:
* poppy-services --http poppy-ergo-jr
* poppy-services --snap poppy-torso
* poppy-services --scratch --vrep poppy-humanoid""",
		formatter_class=RawTextHelpFormatter)

	parser.add_argument('creature', type=str,
						help='poppy creature name',
						action='store', nargs='?',
						choices=list(installed_poppy_creatures.keys()))
	parser.add_argument('--dummy',
						help='use a simulated dummy robot',
						action='store_true')
	parser.add_argument('--vrep',
						help='use a V-REP simulated Poppy Creature',
						action='store_true')
	parser.add_argument('--poppy-simu',
						help='start a simulated dummy robot and the HTTP API to connect to the viewer on simu.poppy-project.org',
						action='store_true')
	parser.add_argument('--snap',
						help='start a Snap! robot server',
						action='store_true')
	parser.add_argument('--snap-port',
						help='port used by the Snap! server',
						default=6969, type=int)
	parser.add_argument('--scratch',
						help='start a Scratch robot server',
						action='store_true')
	parser.add_argument('--scratch-port',
						help='port used by the Scratch server',
						default=8601, type=int)
	parser.add_argument('-nb', '--no-browser',
						help='avoid automatic start of Snap! in web browser',
						action='store_true')
	parser.add_argument('--http',
						help='start a http robot server',
						action='store_true')
	parser.add_argument('--http-port',
						help='port of HttpRobotServer, used for poppy-simu',
						default=8080, type=int)
	parser.add_argument('--remote',
						help='start a remote robot server',
						action='store_true')
	parser.add_argument('--ws',
						help='start the websocket server',
						action='store_true')
	parser.add_argument('--ws-port',
						help='port of Websocket Server',
						default=9009, type=int)
	parser.add_argument('--disable-camera',
						help='Start the robot without the camera.',
						action='store_true')
	parser.add_argument('-v', '--verbose',
						help='start services with verbose mode. There is 3 debug levels, add as "v" as debug level you want',
						action='count')
	parser.add_argument('-f', '--log-file',
						help='Log filename',
						action='store')

	nb_creatures = len(installed_poppy_creatures.keys())
	if nb_creatures == 0:
		print('No installed poppy creature were found!')
		print('You should first install the python package '
			  'corresponding to your robot or check your python environment.')
		sys.exit(1)

	args = parser.parse_args()

	# If no creature are specified and only one is installed
	# We use it as default.
	if args.creature is None:
		if nb_creatures > 1:
			parser.print_help()
			sys.exit(1)

		args.creature = list(installed_poppy_creatures.keys())[0]
		print('No creature specified, use {}'.format(args.creature))

	if args.log_file:
		fh = logging.FileHandler(args.log_file)
		fh.setLevel(logging.DEBUG)
		formatter = logging.Formatter(
			'%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		fh.setFormatter(formatter)
		logging.getLogger('').addHandler(fh)

	if args.verbose:
		args.snap_quiet = False
		args.http_quiet = False
		args.ws_quiet = False

		if args.verbose == 1:
			lvl = logging.WARNING
		elif args.verbose == 2:
			lvl = logging.INFO
		elif args.verbose > 2:
			lvl = logging.DEBUG

		if args.log_file is not None:
			ch = logging.FileHandler(args.log_file)
		else:
			ch = logging.StreamHandler()

		ch.setLevel(lvl)
		formatter = logging.Formatter(
			'%(name)-12s: %(levelname)-8s %(message)s')
		ch.setFormatter(formatter)
		logging.getLogger('').addHandler(ch)

	if not any([args.snap, args.scratch, args.http, args.remote, args.poppy_simu, args.ws, args.dummy]):
		print('No service specified! See --help for details.')
		sys.exit(1)

	scratch_static_server_started = False
	if args.scratch and not args.no_browser:
		scratch_static_port = args.scratch_port
		scratch_static_server = HTTPServer(("0.0.0.0", scratch_static_port), SimpleHTTPRequestHandler)

		from pypot.vpl.scratch import download_scratch_interactively
		static_app = download_scratch_interactively()
		if static_app is None:
			print("The static server was not started because the VPL app has not been downloaded")
		else:
			os.chdir(static_app)
			scratch_static_server_process = Process(target=scratch_static_server.serve_forever, args=())
			scratch_static_server_started = True
			scratch_static_server_process.start()

			scratch_url = 'http://127.0.0.1:{}/'.format(scratch_static_port)
			url = '{}'.format(scratch_url)

	snap_static_server_started = False
	if args.snap and not args.no_browser:
		snap_static_port = 8888
		snap_static_server = HTTPServer(("0.0.0.0", snap_static_port), SimpleHTTPRequestHandler)

		from pypot.vpl.snap import download_snap_interactively
		static_app = download_snap_interactively()
		if static_app is None:
			print("The static server was not started because the VPL app has not been downloaded")
		else:
			os.chdir(static_app)
			snap_static_server_process = Process(target=snap_static_server.serve_forever, args=())
			snap_static_server_started = True
			snap_static_server_process.start()

			snap_url = 'http://127.0.0.1:{}/snap.html'.format(snap_static_port)
			block_url = 'http://{}:{}/snap-blocks.xml'.format(
				find_local_ip(), args.snap_port)
			url = '{}#open:{}'.format(snap_url, block_url)

	with closing(start_poppy_with_services(args)):

		msg = ''

		if args.dummy or args.poppy_simu:
			msg += 'Simulated robot created! He is running on: ip={}'.format(find_local_ip())
		else:
			msg += 'Robot instantiated! He is running on: ip={},'.format(find_local_ip())
			if args.disable_camera:
				msg += ' without camera access.'
			else:
				msg += ' with camera access.'

		if args.vrep: msg += ' With V-REP link.'

		if args.snap or args.scratch or args.ws or args.http or args.poppy_simu:
			msg += '\nServer started on:'
			if args.http or args.scratch or args.poppy_simu: msg += ' http_port={},'.format(args.http_port)
			if args.snap: msg += ' Snap_port={},'.format(args.snap_port)
			if args.ws: msg += ' ws_port={},'.format(args.ws_port)
			msg = msg[0:-1] + '.'

		print(msg)

		sys.stdout.flush()

		if snap_static_server_started or scratch_static_server_started:
			for browser_name in ['firefox', 'chromium-browser', 'chromium', 'google-chrome',
								 'chrome', 'safari', 'midori', None]:
				try:
					browser = webbrowser.get(browser_name)
					browser.open(url, new=0, autoraise=True)
					break
				except Exception:
					pass

		# Just run4ever (until Ctrl-c...)
		try:
			while (True):
				time.sleep(1000)
		except KeyboardInterrupt:
			print("Bye bye!")
			if scratch_static_server_started:
				scratch_static_server_process.terminate()
				scratch_static_server_process.join()
			if snap_static_server_started:
				snap_static_server_process.terminate()
				snap_static_server_process.join()


if __name__ == '__main__':
	main()



import numpy
import logging

from collections import defaultdict

import pypot.utils.pypot_time as time

from ..robot.motor import Motor

from ..utils import SyncEvent
from ..utils.trajectory import GotoMinJerk, GotoLinear
from ..utils.stoppablethread import StoppableLoopThread


logger = logging.getLogger(__name__)


class DxlRegister(object):
    def __init__(self, rw=False):
        self.rw = rw

    def __get__(self, instance, owner):
        if instance._read_synchronous[self.label]:
            sync = instance._read_synced[self.label]

            if not sync.is_recent:
                sync.request()

        value = instance.__dict__.get(self.label, 0)

        return value

    def __set__(self, instance, value):
        if not self.rw:
            raise AttributeError("can't set attribute")

        logger.debug("Setting '%s.%s' to %s",
                     instance.name, self.label, value)
        instance.__dict__[self.label] = value

        if instance._write_synchronous[self.label]:
            sync = instance._write_synced[self.label]
            sync.request()


class DxlOrientedRegister(DxlRegister):
    def __get__(self, instance, owner):
        value = DxlRegister.__get__(self, instance, owner)
        return (value if instance.direct else -value)

    def __set__(self, instance, value):
        value = value if instance.direct else -value
        DxlRegister.__set__(self, instance, value)


class DxlPositionRegister(DxlOrientedRegister):
    def __get__(self, instance, owner):
        value = DxlOrientedRegister.__get__(self, instance, owner)
        return value - instance.offset

    def __set__(self, instance, value):
        value = value + instance.offset
        DxlOrientedRegister.__set__(self, instance, value)


class RegisterOwner(type):
    def __new__(cls, name, bases, attrs):
        for n, v in attrs.items():
            if isinstance(v, DxlRegister):
                v.label = n
                attrs['registers'].append(n)
        return super(RegisterOwner, cls).__new__(cls, name, bases, attrs)


class DxlMotor(Motor, metaclass=RegisterOwner):
    """ High-level class used to represent and control a generic dynamixel motor.

        This class provides all level access to (see :attr:`~pypot.dynamixel.motor.DxlMotor.registers` for an exhaustive list):
            * motor id
            * motor name
            * motor model
            * present position/speed/load
            * goal position/speed/load
            * compliant
            * motor orientation and offset
            * angle limit
            * temperature
            * voltage

        This class represents a generic robotis motor and you define your own subclass for specific motors (see :class:`~pypot.dynamixel.motor.DxlMXMotor` or :class:`~pypot.dynamixel.motor.DxlAXRXMotor`).

        Those properties are synchronized with the real motors values thanks to a :class:`~pypot.dynamixel.controller.DxlController`.

        """

    registers = Motor.registers + ['registers',
                                   'goal_speed',
                                   'compliant', 'safe_compliant',
                                   'angle_limit']

    id = DxlRegister()
    name = DxlRegister()
    model = DxlRegister()

    present_position = DxlPositionRegister()
    goal_position = DxlPositionRegister(rw=True)
    present_speed = DxlOrientedRegister()
    moving_speed = DxlOrientedRegister(rw=True)
    present_load = DxlOrientedRegister()
    torque_limit = DxlRegister(rw=True)

    lower_limit = DxlPositionRegister()
    upper_limit = DxlPositionRegister()
    present_voltage = DxlRegister()
    present_temperature = DxlRegister()

    def __init__(self, id, name=None, model='',
                 direct=True, offset=0.0,
                 broken=False,
                 angle_limit=None):
        self.__dict__['id'] = id

        name = name if name is not None else 'motor_{}'.format(id)
        self.__dict__['name'] = name

        self.__dict__['model'] = model
        self.__dict__['direct'] = direct
        self.__dict__['offset'] = offset

        self.__dict__['compliant'] = True

        self._safe_compliance = SafeCompliance(self)
        self.goto_behavior = 'dummy'
        self.compliant_behavior = 'dummy'

        self._broken = broken

        self._read_synchronous = defaultdict(lambda: False)
        self._read_synced = defaultdict(SyncEvent)

        self._write_synchronous = defaultdict(lambda: False)
        self._write_synced = defaultdict(SyncEvent)

        if angle_limit is not None:
            self.__dict__['lower_limit'], self.__dict__['upper_limit'] = angle_limit

    def __repr__(self):
        return ('<DxlMotor name={self.name} '
                'id={self.id} '
                'pos={self.present_position}>').format(self=self)

    @property
    def goal_speed(self):
        """ Goal speed (in degrees per second) of the motor.

            This property can be used to control your motor in speed. Setting a goal speed will automatically change the moving speed and sets the goal position as the angle limit.

            .. note:: The motor will turn until reaching the angle limit. But this is not a wheel mode, so the motor will stop at its limits.

            """
        return numpy.sign(self.goal_position) * self.moving_speed

    @goal_speed.setter
    def goal_speed(self, value):
        if abs(value) < numpy.finfo(numpy.float).eps:
            self.goal_position = self.present_position

        else:
            # 0.7 corresponds approx. to the min speed that will be converted into 0
            # and as 0 corresponds to setting the max speed, we have to check this case
            value = numpy.sign(value) * 0.7 if abs(value) < 0.7 else value

            self.goal_position = numpy.sign(value) * self.max_pos
            self.moving_speed = abs(value)

    @property
    def compliant_behavior(self):
        return self._compliant_behavior

    @compliant_behavior.setter
    def compliant_behavior(self, value):
        if value not in ('dummy', 'safe'):
            raise ValueError('Wrong compliant type! It should be either "dummy" or "safe".')

        if hasattr(self, '_compliant_behavior') and self._compliant_behavior == value:
            return

        self._compliant_behavior = value

        # Start the safe compliance behavior when the motor should be compliant
        if value == 'safe' and self.compliant:
            self._safe_compliance.start()

        if value == 'dummy':
            use_safe = self._safe_compliance.started
            if use_safe:
                self._safe_compliance.stop()
            self.compliant = self.compliant or use_safe

    @property
    def compliant(self):
        return bool(self.__dict__['compliant'])

    @compliant.setter
    def compliant(self, is_compliant):
        if self._safe_compliance.started and is_compliant:
            return

        if self.compliant_behavior == 'dummy':
            self._set_compliancy(is_compliant)

        elif self.compliant_behavior == 'safe':
            if is_compliant:
                self._safe_compliance.start()
            elif self._safe_compliance.started:
                self._safe_compliance.stop()

    def _set_compliancy(self, is_compliant):
        # Change the goal_position only if you switch from compliant to not compliant mode
        if not is_compliant and self.compliant:
            self.goal_position = self.present_position
        self.__dict__['compliant'] = is_compliant

    @property
    def angle_limit(self):
        return self.lower_limit, self.upper_limit

    @angle_limit.setter
    def angle_limit(self, limits):
        self.lower_limit, self.upper_limit = limits

    @property
    def goto_behavior(self):
        return self._default_goto_behavior

    @goto_behavior.setter
    def goto_behavior(self, value):
        if value not in ('dummy', 'minjerk', 'linear'):
            raise ValueError('Wrong compliant type! It should be either "dummy", "minjerk" or "linear".')
        self._default_goto_behavior = value

    def goto_position(self, position, duration, control=None, wait=False):
        """ Automatically sets the goal position and the moving speed to reach the desired position within the duration. """

        if control is None:
            control = self.goto_behavior

        if control == 'minjerk':
            goto_min_jerk = GotoMinJerk(self, position, duration)
            goto_min_jerk.start()

            if wait:
                goto_min_jerk.wait_to_stop()

        elif control == 'dummy':
            dp = abs(self.present_position - position)
            speed = (dp / float(duration)) if duration > 0 else 0

            self.moving_speed = speed
            self.goal_position = position

            if wait:
                time.sleep(duration)

        elif control == 'linear':
            goto_linear = GotoLinear(self, position, duration)
            goto_linear.start()

            if wait:
                goto_linear.wait_to_stop()


class DxlAXRXMotor(DxlMotor):
    """ This class represents the AX robotis motor.

        This class adds access to:
            * compliance margin/slope (see the robotis website for details)

        """
    registers = list(DxlMotor.registers)

    compliance_margin = DxlRegister(rw=True)
    compliance_slope = DxlRegister(rw=True)

    def __init__(self, id, name=None, model='',
                 direct=True, offset=0.0, broken=False,
                 angle_limit=None):
        DxlMotor.__init__(self, id, name, model,
                          direct, offset, broken,
                          angle_limit)
        self.max_pos = 150


class DxlMXMotor(DxlMotor):
    """ This class represents the RX and MX robotis motor.

        This class adds access to:
            * PID gains (see the robotis website for details)

        """
    registers = list(DxlMotor.registers)

    pid = DxlRegister(rw=True)

    def __init__(self, id, name=None, model='',
                 direct=True, offset=0.0, broken=False,
                 angle_limit=None):
        """ This class represents the RX and MX robotis motor.

            This class adds access to:
                * PID gains (see the robotis website for details)

            """
        DxlMotor.__init__(self, id, name, model,
                          direct, offset, broken,
                          angle_limit)
        self.max_pos = 180


class DxlMX64106Motor(DxlMXMotor):
    """ This class represents the MX-64 and MX-106 robotis motor.

        This class adds access to:
            * present current

        """

    registers = list(DxlMXMotor.registers)

    present_current = DxlRegister()

    def __init__(self, id, name=None, model='',
                 direct=True, offset=0.0, broken=False,
                 angle_limit=None):
        """ This class represents the RX and MX robotis motor.

            This class adds access to:
                * PID gains (see the robotis website for details)

            """
        DxlMotor.__init__(self, id, name, model,
                          direct, offset, broken,
                          angle_limit)
        self.max_pos = 180


class DxlXL320Motor(DxlMXMotor):
    registers = list(DxlMXMotor.registers)

    led = DxlRegister(rw=True)
    control_mode = DxlRegister(rw=True)

    """ This class represents the XL-320 robotis motor. """
    def __init__(self, id, name=None, model='XL-320',
                 direct=True, offset=0.0, broken=False,
                 angle_limit=None):
        DxlMXMotor.__init__(self, id, name, model,
                            direct, offset, broken,
                            angle_limit)
        self.max_pos = 150


class DxlSRMotor(DxlMotor):
    """ This class represents the robotis motor found in the seed robotics hand.

        This class adds access to:
            * force control enable
            * goal force
            * present current

        """
    registers = list(DxlMotor.registers)

    force_control_enable = DxlRegister(rw=True)
    goal_force = DxlRegister(rw=True)
    present_current = DxlRegister()

    def __init__(self, id, name=None, model='',
                 direct=True, offset=0.0, broken=False,
                 angle_limit=None):
        """ This class represents the robotis motor found in the seed robotics hand.

        This class adds access to:
            * PID gains (see the robotis website for details)
            * force control enable
            * goal force

            """
        DxlMotor.__init__(self, id, name, model,
                          direct, offset, broken,
                          angle_limit)
        self.max_pos = 180


class SafeCompliance(StoppableLoopThread):
    """ This class creates a controller to active compliance only if the current motor position is included in the angle limit, else the compliance is turned off. """

    def __init__(self, motor, frequency=50):
        StoppableLoopThread.__init__(self, frequency)

        self.motor = motor

    def update(self):
        self.motor._set_compliancy((min(self.motor.angle_limit) < self.motor.present_position < max(self.motor.angle_limit)))

    def teardown(self):
        self.motor._set_compliancy(False)



from scipy.spatial import cKDTree
from scipy.interpolate import interp1d
import numpy as np


class KDTreeDict(dict):

    def __init__(self, gen_tree_on_add=False, distance_upper_bound=0.2, k_neighbors=2):
        super(KDTreeDict, self).__init__()
        self.gen_tree_on_add = gen_tree_on_add
        self.distance_upper_bound = distance_upper_bound
        self.k_neighbors = k_neighbors
        self.__keys = []
        self.__tree = None
        self.__stale = False

    def __setitem__(self, key, val):

        if not isinstance(key, tuple):
            _key = (key,)
        if _key not in self.__keys:
            self.__keys.append(_key)
            self.__stale = True
            if self.gen_tree_on_add:
                self.generate_tree()
        super(KDTreeDict, self).__setitem__(key, val)

    def __getitem__(self, key):
        # if not isinstance(key, tuple):
        #     key = (key,)
        return super(KDTreeDict, self).__getitem__(key)


    def __len__(self):
        return len(self.__keys)

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))
        other = dict(*args, **kwargs)
        for key in other:
            self[key] = other[key]

    def generate_tree(self):
        self.__tree = cKDTree(self.__keys)
        self.__stale = False

    def nearest_keys(self, key):
        """Find the nearest_keys (l2 distance) as strings thanks to a cKDTree query"""
        if not isinstance(key, tuple):
            _key = (key,)
        if self.__stale:
            self.generate_tree()
        d, idx = self.__tree.query(
            _key, self.k_neighbors, distance_upper_bound=self.distance_upper_bound)

        try:
            return [self.__keys[id][0] for id in idx if id < len(self.__keys)]
        except TypeError:
            # if k_neighbors = 1 query is not returnng arrays
            return self.__keys[idx]

    def interpolate_motor_positions(self, input_key, nearest_keys):
        """
            Process linear interpolation to estimate current speed and position of motors
            Method specific to the :meth:~pypot.primitive.move.Move.position() structure
            it is a KDTreeDict[timestamp] = {dict[motor]=(position,speed)}

        :param input_key: string of a float as returned by cKDTree.query()
        :param nearest_keys: 2-tuples of strings of a float as returned by cKDTree.query()
        """

        # WARNING: input_key and nearest_keys are string encoding floats as returned by cKDTree
        # TODO: to be rewrited with more style (map ?)

        if len(nearest_keys) == 1:
            return self[nearest_keys[0]]
        elif len(nearest_keys) == 0:
            raise KeyError('key {} exceed distance_upper_bound {}'.format(
                input_key, self.distance_upper_bound))
        elif len(nearest_keys) != 2:
            raise NotImplementedError("interpolation works only for k_neighbors = 2")
        elif nearest_keys[0] == nearest_keys[1]:
            # Bug from nearest key ?
            return self[nearest_keys[0]]
        # Problem if ValueError: A value in x_new is above the interpolation range.
        else:
            float_nearest_keys = [float(k) for k in nearest_keys]
            float_input_key = float(input_key)
            if float_input_key < min(float_nearest_keys):
                return self[min(float_nearest_keys)]
            elif float_input_key > max(float_nearest_keys):
                return self[max(float_nearest_keys)]
            else:              
                interpolated_positions = {}
                for (k, v), (k2, v2) in zip(self[nearest_keys[0]].items(), self[nearest_keys[1]].items()):
                    if k == k2:
                        x = np.array(float_nearest_keys)
                        y_pos = np.array([v[0], v2[0]])
                        y_speed = np.array([v[1], v2[1]])
                        f_pos = interp1d(x, y_pos, bounds_error=False)
                        f_speed = interp1d(x, y_speed, bounds_error=False)

                        pos = f_pos(float_input_key)
                        speed = f_speed(float_input_key)
                        interpolated_positions[k] = (pos, speed)
                    else:
                        raise IndexError("key are not identics. Motor added during the record ?")
                return interpolated_positions

    def __missing__(self, key):
        if key is None:
            raise SyntaxError('invalid syntax, you must provide a key')
        return self.interpolate_motor_positions(key, self.nearest_keys(key))



try:
    from .httpserver import HTTPRobotServer
except ImportError:
    pass

try:
    from .zmqserver import ZMQRobotServer
except ImportError:
    pass

try:
    from .server import RemoteRobotServer
except ImportError:
    pass

try:
    from .ws import WsRobotServer
except ImportError:
    pass



import logging

from ..primitive.manager import PrimitiveManager


logger = logging.getLogger(__name__)


class Robot(object):
    """ This class is used to regroup all motors and sensors of your robots.

    Most of the time, you do not want to directly instantiate this class, but you rather want to use a factory which creates a robot instance - e.g. from a python dictionnary (see :ref:`config_file`).

    This class encapsulates the different controllers (such as dynamixel ones) that automatically synchronize the virtual sensors/effectors instances held by the robot class with the real devices. By doing so, each sensor/effector can be synchronized at a different frequency.

    This class also provides a generic motors accessor in order to (more or less) easily extends this class to other types of motor.

        """
    def __init__(self, motor_controllers=[], sensor_controllers=[], sync=True):
        """
        :param list motor_controllers: motors controllers to attach to the robot
        :param list sensor_controllers: sensors controllers to attach to the robot
        :param bool sync: choose if automatically starts the synchronization loops

        """
        self._motors = []
        self._sensors = []
        self.alias = []

        self._controllers = sensor_controllers + motor_controllers

        for controller in motor_controllers:
            for m in controller.motors:
                setattr(self, m.name, m)

            self._motors.extend(controller.motors)

        for controller in sensor_controllers:
            for s in controller.sensors:
                setattr(self, s.name, s)

            self._sensors.extend(controller.sensors)

        self._attached_primitives = {}
        self._primitive_manager = PrimitiveManager(self.motors)

        self._syncing = False
        if sync:
            self.start_sync()

    def close(self):
        """ Cleans the robot by stopping synchronization and all controllers."""
        self.stop_sync()
        [c.io.close() for c in self._controllers if c.io is not None]

    def __repr__(self):
        return '<Robot motors={}>'.format(self.motors)

    def start_sync(self):
        """ Starts all the synchonization loop (sensor/effector controllers). """
        if self._syncing:
            return

        [c.start() for c in self._controllers]
        [c.wait_to_start() for c in self._controllers]
        self._primitive_manager.start()
        self._primitive_manager._running.wait()

        self._syncing = True

        logger.info('Starting robot synchronization.')

    def stop_sync(self):
        """ Stops all the synchonization loop (sensor/effector controllers). """
        if not self._syncing:
            return

        if self._primitive_manager.running:
            self._primitive_manager.stop()

        [c.stop() for c in self._controllers]
        [s.close() for s in self.sensors if hasattr(s, 'close')]

        self._syncing = False

        logger.info('Stopping robot synchronization.')

    def attach_primitive(self, primitive, name):
        setattr(self, name, primitive)
        self._attached_primitives[name] = primitive
        primitive.name = name

        logger.info("Attaching primitive '%s' to the robot.", name)

    @property
    def motors(self):
        """ Returns all the motors attached to the robot. """
        return self._motors

    @property
    def sensors(self):
        """ Returns all the sensors attached to the robot. """
        return self._sensors

    @property
    def active_primitives(self):
        """ Returns all the primitives currently running on the robot. """
        return self._primitive_manager._prim

    @property
    def primitives(self):
        """ Returns all the primitives name attached to the robot. """
        return list(self._attached_primitives.values())

    @property
    def compliant(self):
        """ Returns a list of all the compliant motors. """
        return [m for m in self.motors if m.compliant]

    @compliant.setter
    def compliant(self, is_compliant):
        """ Switches all motors to compliant (resp. non compliant) mode. """
        for m in self.motors:
            m.compliant = is_compliant

    def goto_position(self, position_for_motors, duration, control=None, wait=False):
        """ Moves a subset of the motors to a position within a specific duration.

            :param dict position_for_motors: which motors you want to move {motor_name: pos, motor_name: pos,...}
            :param float duration: duration of the move
            :param str control: control type ('dummy', 'minjerk')
            :param bool wait: whether or not to wait for the end of the move

            .. note::In case of dynamixel motors, the speed is automatically adjusted so the goal position is reached after the chosen duration.

            """
        for i, (motor_name, position) in enumerate(position_for_motors.items()):
            w = False if i < len(position_for_motors) - 1 else wait

            m = getattr(self, motor_name)
            m.goto_position(position, duration, control, wait=w)

    def power_up(self):
        """ Changes all settings to guarantee the motors will be used at their maximum power. """
        for m in self.motors:
            m.compliant = False
            m.moving_speed = 0
            m.torque_limit = 100.0

    def to_config(self):
        """ Generates the config for the current robot.

            .. note:: The generated config should be used as a basis and must probably be modified.

        """
        from ..dynamixel.controller import DxlController

        dxl_controllers = [c for c in self._controllers
                           if isinstance(c, DxlController)]

        config = {}

        config['controllers'] = {}
        for i, c in enumerate(dxl_controllers):
            name = 'dxl_controller_{}'.format(i)
            config['controllers'][name] = {
                'port': c.io.port,
                'sync_read': c.io._sync_read,
                'attached_motors': [m.name for m in c.motors],
            }

        config['motors'] = {}
        for m in self.motors:
            config['motors'][m.name] = {
                'id': m.id,
                'type': m.model,
                'offset': m.offset,
                'orientation': 'direct' if m.direct else 'indirect',
                'angle_limit': m.angle_limit,
            }

            if m.angle_limit == (0, 0):
                config['motors']['wheel_mode'] = True

        config['motorgroups'] = {}

        return config





import time
import numpy

from collections import deque

from ...robot.sensor import Sensor
from ...utils import StoppableLoopThread
from ...utils.i2c_controller import I2cController


class SonarSensor(Sensor):
    """ Give access to ultrasonic I2C modules SRF-02 in a *pypot way*

        It provides one register: distance (in meters).

    """
    registers = Sensor.registers + ['distance']

    def __init__(self, name, i2c_pin, address, sync_freq=50.0):
        Sensor.__init__(self, name)

        self._d = numpy.nan

        self._sonar = Sonar(i2c_pin, [address])

        self._controller = StoppableLoopThread(sync_freq, update=self.update)
        self._controller.start()

    def close(self):
        self._controller.stop()

    def update(self):
        self._sonar.update()
        self.distance = self._sonar.data[0]

    @property
    def distance(self):
        return self._d

    @distance.setter
    def distance(self, d):
        self._d = d / 100


class Sonar(object):
    """ Give access to ultrasonic I2C modules SRF-02 connected with I2C pin of your board.
        To get more information, go to http://www.robot-electronics.co.uk/htm/srf02techI2C.htm

        Example:

        > i2c = smbus.SMBus(1)
        > sonar = Sonar(i2c, addresses=[0x70, 0x71, 0x72])
        >
    """

    def __init__(self, pin_number, addresses=[0x70]):
        """ 0x70 is the default address for the SRF-02 I2C module. """

        self.i2c = I2cController(pin_number)
        self.addresses = addresses

        self.data = None

        self._raw_data_queues = [deque([], 5) for _ in addresses]

        self.results_type = {'inches': 0x50,
                             'centimeters': 0x51,
                             'microseconds': 0x52}

        self.__errors = 0

    def update(self):
        self.ping()
        time.sleep(0.065)
        self.data = self._filter(self.read())
        return self.data

    def ping(self):
        for addr in self.addresses:
            self._ping(addr)

    def read(self, reg=2):
        return [self._read(addr, reg) for addr in self.addresses]

    def _filter(self, data):
        """ Apply a filter to reduce noisy data.

           Return the median value of a heap of data.

        """
        filtered_data = []
        for queue, data in zip(self._raw_data_queues, data):
            queue.append(data)
            filtered_data.append(numpy.median(queue))

        return filtered_data

    def _ping(self, address, data=None):
        d = data if data is not None else self.results_type['centimeters']

        while True:
            try:
                self.i2c.write_byte_data(address, 0, d)
                break
            except IOError:
                time.sleep(0.005)
                self.__errors += 1

    def _read(self, address, reg=2):
        while True:
            try:
                return int(self.i2c.read_word_data(address, reg)) / 256
            except IOError:
                time.sleep(0.005)
                self.__errors += 1


if __name__ == '__main__':
    import smbus

    from pylab import plot

    i2c = smbus.SMBus(1)
    sonar = Sonar(i2c)

    d = []
    t = [time.time()]
    for _ in range(1000):
        sonar.update()
        d.append(sonar.data[0])
        t.append(time.time() - t[0])
    plot(t[1:], d)



import time
import numpy
import datetime
import threading

from collections import namedtuple


TrackedObject = namedtuple('TrackedObject', ('position', 'quaternion', 'orientation', 'timestamp'))


def quat2euler(q):
    qx, qy, qz, qw = q
    sqx, sqy, sqz, sqw = q ** 2
    invs = 1.0 / (sqx + sqy + sqz + sqw)

    yaw = numpy.arctan2(2.0 * (qx * qz + qy * qw) * invs, (sqx - sqy - sqz + sqw) * invs)
    pitch = -numpy.arcsin(2.0 * (qx * qy - qz * qw) * invs)
    roll = numpy.arctan2(2.0 * (qy * qz + qx * qw) * invs, (-sqx + sqy - sqz + sqw) * invs)

    return numpy.array((yaw, pitch, roll))


try:
    import vrpn

    class OptiTrackClient(threading.Thread):
        """ Retrieves position, orientation, and timestamp of each tracked object.

            The position is expressed in meters (X is left, Y is up, and Z is depth).
            The orientation is expressed in radians (yaw, pitch, roll).

            """
        def __init__(self, addr, port, obj_names):
            threading.Thread.__init__(self)
            self.daemon = True

            self.trackers = []
            for obj in obj_names:
                t = vrpn.receiver.Tracker('{}@{}:{}'.format(obj, addr, port))
                t.register_change_handler(obj, self.handler, 'position')
                self.trackers.append(t)

            self._tracked_objects = {}

        @property
        def tracked_objects(self):
            return self._tracked_objects

        @property
        def recent_tracked_objects(self):
            """ Only returns the objects that have been tracked less than 20ms ago. """
            dt = 0.02
            f = lambda name: (datetime.datetime.now() - self.tracked_objects[name].timestamp).total_seconds()
            return dict([(k, v) for k, v in self.tracked_objects.items() if f(k) < dt])

        def handler(self, obj, data):
            self.tracked_objects[obj] = TrackedObject(numpy.array(*data['position']),
                                                      numpy.array(data['quaternion']),
                                                      quat2euler(numpy.array(data['quaternion'])),
                                                      datetime.datetime.now())

        def serve_forever(self):
            self.start()

            while True:
                try:
                    self.join(timeout=1.0)
                except KeyboardInterrupt:
                    break

        def run(self):
            while True:
                for t in self.trackers:
                    t.mainloop()
                    time.sleep(1.0 / 120)

except ImportError:
    pass

if __name__ == '__main__':
    c = OptiTrackClient('193.50.110.176', 3883, ('obj_1', ))
    c.serve_forever()



#!/usr/bin/env python



"""
Poppy configuration tools

Examples:
* poppy-configure ergo-jr m2

"""
import sys

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from subprocess import call

from pypot.creatures import installed_poppy_creatures
from pypot.utils import flushed_print as print


def find_port_for_motor(config, motor):
    def flatten_motorgroups(motors, groups):
        l = []
        for m in motors:
            if m in groups:
                l += flatten_motorgroups(groups[m], groups)
            else:
                l.append(m)
        return l
    
    for bus in config['controllers']:
        motors_on_bus = config['controllers'][bus]["attached_motors"]
        motors = flatten_motorgroups(motors_on_bus, config['motorgroups'])
        if motor in motors:
            return config['controllers'][bus]["port"]

    raise ValueError('Something must be wrong in your configuration file. '
                     'Could not find bus for motor {}'.format(motor))


def main():
    robots = [c.replace('poppy-', '') for c in installed_poppy_creatures]

    parser = ArgumentParser(description='Configuration tool for Poppy robots ',
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('robot', type=str, choices=robots,
                        help='Robot used.')

    parser.add_argument('motor', type=str,
                        help='Name of the motor to configure.')

    args = parser.parse_args()

    RobotCls = installed_poppy_creatures['poppy-{}'.format(args.robot)]
    c = RobotCls.default_config

    if args.motor not in c['motors']:
        print('"{}" is not a motor of "{}"! '
              'possibilities={}'.format(args.motor, args.robot,
                                        sorted(c['motors'].keys())))
        print('Exiting now...')
        sys.exit(1)

    motor_config = c['motors'][args.motor]

    args = [
        '--id', motor_config['id'],
        '--type', motor_config['type'],
        '--port', find_port_for_motor(c, args.motor),
        '--return-delay-time', 0
    ]
    
    if 'wheel_mode' in motor_config.keys():
        args.extend(('--wheel-mode', motor_config['wheel_mode']))
    else:
        args.extend(('--angle-limit',motor_config['angle_limit'][0],motor_config['angle_limit'][1],
                     '--goto-zero'))
        
    call(['dxl-config'] + list(map(str, args)))


if __name__ == '__main__':
    main()



from itertools import repeat

from .abstract_io import (AbstractDxlIO, _DxlControl, _DxlAccess,
                          DxlTimeoutError, DxlCommunicationError)
from .. import conversion as conv
from ..protocol import v2 as v2


class Dxl320IO(AbstractDxlIO):
    _protocol = v2

    def set_wheel_mode(self, ids):
        self.set_control_mode(dict(zip(ids, repeat('wheel'))))

    def set_joint_mode(self, ids):
        self.set_control_mode(dict(zip(ids, repeat('joint'))))

    def get_goal_position_speed_load(self, ids):
        a = self._get_goal_pos_speed(ids)
        b = self.get_torque_limit(ids)

        return list(zip(*list(zip(*a)) + [b]))

    def set_goal_position_speed_load(self, value_for_ids):
        values = list(zip(*list(value_for_ids.values())))

        self._set_goal_pos_speed(dict(zip(value_for_ids.keys(),
                                          zip(*(values[0], values[1])))))

        self.set_torque_limit(dict(zip(value_for_ids.keys(), values[2])))

    def factory_reset(self, ids, except_ids=False, except_baudrate_and_ids=False):
        """ Reset all motors on the bus to their factory default settings. """

        mode = (0x02 if except_baudrate_and_ids else
                0x01 if except_ids else 0xFF)

        for id in ids:
            try:
                self._send_packet(self._protocol.DxlResetPacket(id, mode))

            except (DxlTimeoutError, DxlCommunicationError):
                pass

# TODO:
#   * error

# MARK: - Generate the accessors


controls = {
    # EEPROM
    'model': {
        'address': 0x00,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_model
    },
    'firmware': {
        'address': 0x02,
        'length': 1,
        'access': _DxlAccess.readonly
    },
    'id': {
        'address': 0x03,
        'length': 1,
        'access': _DxlAccess.writeonly,
        'setter_name': 'change_id'
    },
    'baudrate': {
        'address': 0x04,
        'length': 1,
        'access': _DxlAccess.writeonly,
        'setter_name': 'change_baudrate',
        'si_to_dxl': conv.baudrate_to_dxl
    },
    'return delay time': {
        'address': 0x05,
        'length': 1,
        'dxl_to_si': conv.dxl_to_rdt,
        'si_to_dxl': conv.rdt_to_dxl
    },
    'angle limit': {
        'address': 0x06,
        'nb_elem': 2,
        'dxl_to_si': lambda value, model: (conv.dxl_to_degree(value[0], model),
                                           conv.dxl_to_degree(value[1], model)),
        'si_to_dxl': lambda value, model: (conv.degree_to_dxl(value[0], model),
                                           conv.degree_to_dxl(value[1], model))
    },
    'control mode': {
        'address': 0x0B,
        'length': 1,
        'dxl_to_si': conv.dxl_to_control_mode,
        'si_to_dxl': conv.control_mode_to_dxl,
    },
    'highest temperature limit': {
        'address': 0x0C,
        'length': 1,
        'dxl_to_si': conv.dxl_to_temperature,
        'si_to_dxl': conv.temperature_to_dxl
    },
    'voltage limit': {
        'address': 0x0D,
        'length': 1,
        'nb_elem': 2,
        'dxl_to_si': lambda value, model: (conv.dxl_to_voltage(value[0], model),
                                           conv.dxl_to_voltage(value[1], model)),
        'si_to_dxl': lambda value, model: (conv.voltage_to_dxl(value[0], model),
                                           conv.voltage_to_dxl(value[1], model))
    },
    'max torque': {
        'address': 0x0F,
        'dxl_to_si': conv.dxl_to_torque,
        'si_to_dxl': conv.torque_to_dxl
    },
    'status return level': {
        'address': 0x11,
        'length': 1,
        'dxl_to_si': conv.dxl_to_status,
        'si_to_dxl': conv.status_to_dxl
    },
    'alarm shutdown': {
        'address': 0x12,
        'length': 1,
        'dxl_to_si': conv.dxl_to_alarm,
        'si_to_dxl': conv.alarm_to_dxl
    },
    # RAM
    'torque_enable': {
        'address': 0x18,
        'length': 1,
        'dxl_to_si': conv.dxl_to_bool,
        'si_to_dxl': conv.bool_to_dxl,
        'getter_name': 'is_torque_enabled',
        'setter_name': '_set_torque_enable'
    },
    'LED': {
        'address': 0x19,
        'length': 1,
        'dxl_to_si': conv.dxl_to_bool,
        'si_to_dxl': conv.bool_to_dxl,
        'setter_name': '_set_LED',
        'getter_name': 'is_led_on'
    },
    'LED color': {
        'address': 0x19,
        'length': 1,
        'dxl_to_si': conv.dxl_to_led_color,
        'si_to_dxl': conv.led_color_to_dxl,
    },
    'pid gain': {
        'address': 0x1B,
        'length': 1,
        'nb_elem': 3,
        'dxl_to_si': conv.dxl_to_pid,
        'si_to_dxl': conv.pid_to_dxl
    },
    'goal position': {
        'address': 0x1E,
        'dxl_to_si': conv.dxl_to_degree,
        'si_to_dxl': conv.degree_to_dxl
    },
    'moving speed': {
        'address': 0x20,
        'dxl_to_si': conv.dxl_to_speed,
        'si_to_dxl': conv.speed_to_dxl
    },
    'torque limit': {
        'address': 0x23,
        'dxl_to_si': conv.dxl_to_torque,
        'si_to_dxl': conv.torque_to_dxl
    },
    'goal position speed': {
        'address': 0x1E,
        'nb_elem': 2,
        'dxl_to_si': lambda value, model: (conv.dxl_to_degree(value[0], model),
                                           conv.dxl_to_speed(value[1], model)),
        'si_to_dxl': lambda value, model: (conv.degree_to_dxl(value[0], model),
                                           conv.speed_to_dxl(value[1], model)),
        'getter_name': '_get_goal_pos_speed',
        'setter_name': '_set_goal_pos_speed'
    },
    'present position': {
        'address': 0x25,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_degree
    },
    'present speed': {
        'address': 0x27,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_speed
    },
    'present load': {
        'address': 0x29,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_load
    },
    'present position speed load': {
        'address': 0x25,
        'nb_elem': 3,
        'access': _DxlAccess.readonly,
        'dxl_to_si': lambda value, model: (conv.dxl_to_degree(value[0], model),
                                           conv.dxl_to_speed(value[1], model),
                                           conv.dxl_to_load(value[2], model))
    },
    'present voltage': {
        'address': 0x2D,
        'length': 1,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_voltage
    },
    'present temperature': {
        'address': 0x2E,
        'length': 1,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_temperature
    },
    'moving': {
        'address': 0x31,
        'length': 1,
        'access': _DxlAccess.readonly,
        'dxl_to_si': conv.dxl_to_bool,
        'getter_name': 'is_moving'
    }
}


def _add_control(name,
                 address, length=2, nb_elem=1,
                 access=_DxlAccess.readwrite,
                 models=['XL-320', ],
                 dxl_to_si=lambda val, model: val,
                 si_to_dxl=lambda val, model: val,
                 getter_name=None,
                 setter_name=None):

    control = _DxlControl(name,
                          address, length, nb_elem,
                          access,
                          models,
                          dxl_to_si, si_to_dxl,
                          getter_name, setter_name)

    Dxl320IO._generate_accessors(control)


for name, args in controls.items():
    _add_control(name, **args)



import time

from threading import Thread

from ...robot.sensor import Sensor


class AbstractCamera(Sensor):
    registers = Sensor.registers + ['frame', 'resolution', 'fps']

    def __init__(self, name, resolution, fps):
        Sensor.__init__(self, name)

        self._res, self._fps = resolution, fps
        self._last_frame = self._grab_and_process()

        self.running = True
        self._processing = Thread(target=self._process_loop)
        self._processing.daemon = True
        self._processing.start()

    @property
    def frame(self):
        return self._last_frame

    def post_processing(self, image):
        return image

    def grab(self):
        raise NotImplementedError

    def _grab_and_process(self):
        return self.post_processing(self.grab())

    def _process_loop(self):
        period = 1.0 / self.fps

        while self.running:
            self._last_frame = self._grab_and_process()
            time.sleep(period)

    @property
    def resolution(self):
        return list(reversed(self.frame.shape[:2]))

    @property
    def fps(self):
        return self._fps

    def close(self):
        self.running = False
        self._processing.join()



import zmq
import time
import pickle
import threading

from . import optitrack


class OptiBridgeServer(threading.Thread):
    def __init__(self, bridge_host, bridge_port,
                 opti_addr, opti_port, obj_name):
        threading.Thread.__init__(self)
        self.daemon = True

        c = zmq.Context()
        self.s = c.socket(zmq.PUB)
        self.s.bind('tcp://{}:{}'.format(bridge_host, bridge_port))

        self.optitrack = optitrack.OptiTrackClient(opti_addr, opti_port, obj_name)
        self.optitrack.start()

        self.obj_name = obj_name

    def run(self):
        while True:
            self.s.send(pickle.dumps(self.optitrack.recent_tracked_objects))
            time.sleep(0.02)


class OptiTrackClient(threading.Thread):
    def __init__(self, bridge_host, bridge_port, obj_name):
        threading.Thread.__init__(self)
        self.daemon = True

        c = zmq.Context()
        self.s = c.socket(zmq.SUB)
        self.s.connect('tcp://{}:{}'.format(bridge_host, bridge_port))
        self.s.setsockopt(zmq.SUBSCRIBE, '')

        self.obj_name = obj_name
        self._tracked_obj = {}

    def run(self):
        while True:
            d = pickle.loads(self.s.recv())
            self._tracked_obj = {
                k: d[k]
                for k in [k for k in list(d.keys()) if k in self.obj_name]
            }

    @property
    def tracked_objects(self):
        return self._tracked_obj

    @property
    def recent_tracked_objects(self):
        return self.tracked_objects



from .rest import RESTRobot

import logging

logger = logging.getLogger(__name__)


class AbstractServer(object):
    def __init__(self, robot, host, port):
        self.restful_robot = RESTRobot(robot)
        self.host, self.port = host, port

    def run(self):
        raise NotImplementedError


class RemoteRobotServer(AbstractServer):
    def run(self):
        try:
            import zerorpc
            server = zerorpc.Server(self.restful_robot)
            server.bind('tcp://{}:{}'.format(self.host, self.port))
            server.run()
        except ImportError:
            logger.warning(("Warning: The Python module 'zerorpc' is not installed. "
                            "Therefore the feature RemoteRobotServer is disabled. "
                            "On most systems this module can be installed with the command 'pip install zerorpc'."))





import sys
import logging
import json
import os
import re

from threading import Thread

from pypot.robot import Robot, from_json, use_dummy_robot
from pypot.server.snap import SnapRobotServer, find_local_ip

logger = logging.getLogger(__name__)

MAX_SETUP_TRIALS = 10


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


def camelcase_to_underscore(name):
    return re.sub('([a-z])([A-Z0-9])', r'\1_\2', name).lower()


class AbstractPoppyCreature(Robot):
    """ Abstract Class for Any Poppy Creature. """
    def __new__(cls,
                base_path=None, config=None,
                simulator=None, scene=None, host='localhost', port=19997, id=None, shared_vrep_io=None,
                use_snap=False, snap_host='0.0.0.0', snap_port=6969, snap_quiet=True,
                use_http=False, http_host='0.0.0.0', http_port=8080, http_quiet=True,
                use_remote=False, remote_host='0.0.0.0', remote_port=4242,
                use_ws=False, ws_host='0.0.0.0', ws_port=9009,
                start_background_services=True, sync=True,
                **extra):
        """ Poppy Creature Factory.

        Creates a Robot (real or simulated) and specifies it to make it a specific Poppy Creature.

        :param str config: path to a specific json config (if None uses the default config of the poppy creature - e.g. poppy_humanoid.json)

        :param str simulator: name of the simulator used : 'vrep', 'poppy-simu', or 'dummy'
        :param str scene: specify a particular simulation scene (if None uses the default scene of the poppy creature, use "keep-existing" to keep the current VRep scene - e.g. poppy_humanoid.ttt)
        :param str host: host of the simulator
        :param int port: port of the simulator
        :param int id: robot id in simulator (useful when using a scene with multiple robots)
        :param vrep_io: use an already connected VrepIO (useful when using a scene with multiple robots)
        :type vrep_io: :class:`~pypot.vrep.io.VrepIO`
        :param bool use_snap: start or not the Snap! API
        :param str snap_host: host of Snap! API
        :param int snap_port: port of the Snap!
        :param bool use_http: start or not the HTTP API
        :param str http_host: host of HTTP API
        :param int http_port: port of the HTTP API
        :param int id: id of robot in the v-rep scene (not used yet!)
        :param bool sync: choose if automatically starts the synchronization loops

        You can also add extra keyword arguments to disable sensor. For instance, to use a DummyCamera, you can add the argument: camera='dummy'.

        .. warning:: You can not specify a particular config when using a simulated robot!

        """
        if config and simulator:
            raise ValueError('Cannot set a specific config '
                             'when using a simulated version!')

        creature = camelcase_to_underscore(cls.__name__)
        base_path = (os.path.dirname(__import__(creature).__file__)
                     if base_path is None else base_path)

        default_config = os.path.join(os.path.join(base_path, 'configuration'),
                                      '{}.json'.format(creature))

        if config is None:
            config = default_config

        if simulator is not None:
            if simulator == 'vrep':
                from pypot.vrep import from_vrep, VrepConnectionError

                scene_path = os.path.join(base_path, 'vrep-scene')
                if scene != "keep-existing":
                    if scene is None:
                        scene = '{}.ttt'.format(creature)

                    elif not os.path.exists(scene):
                        if ((os.path.basename(scene) != scene) or
                                (not os.path.exists(os.path.join(scene_path, scene)))):
                            raise ValueError('Could not find the scene "{}"!'.format(scene))

                    scene = os.path.join(scene_path, scene)
                # TODO: use the id so we can have multiple poppy creatures
                # inside a single vrep scene

                # vrep.simxStart no longer listen on localhost
                if host == 'localhost':
                    host = '127.0.0.1'

                try:
                    poppy_creature = from_vrep(config, host, port, scene if scene != "keep-existing" else None, id=id, shared_vrep_io=shared_vrep_io)
                except VrepConnectionError:
                    raise IOError('Connection to V-REP failed!')

            elif simulator == 'poppy-simu':
                use_http = True
                poppy_creature = use_dummy_robot(config)
            elif simulator == 'dummy':
                poppy_creature = use_dummy_robot(config)
            else:
                raise ValueError('Unknown simulation mode: "{}"'.format(simulator))

            poppy_creature.simulated = True

        else:
            for _ in range(MAX_SETUP_TRIALS):
                try:
                    poppy_creature = from_json(config, sync, **extra)
                    logger.info('Init successful')
                    break
                except Exception as e:
                    logger.warning('Init fail: {}'.format(str(e)))
                    exc_type, exc_inst, tb = sys.exc_info()

            else:
                import traceback
                traceback.print_tb(tb)
                raise OSError('Could not initialize robot: {}'.format(exc_inst))
            poppy_creature.simulated = False

        with open(config) as f:
            poppy_creature.config = json.load(f)

        urdf_file = os.path.join(os.path.join(base_path,
                                              '{}.urdf'.format(creature)))
        poppy_creature.urdf_file = urdf_file

        if use_snap:
            poppy_creature.snap = SnapRobotServer(
                poppy_creature, snap_host, snap_port, quiet=snap_quiet)
            snap_url = 'http://snap.berkeley.edu/snapsource/snap.html'
            block_url = 'http://{}:{}/snap-blocks.xml'.format(find_local_ip(), snap_port)
            url = '{}#open:{}'.format(snap_url, block_url)
            logger.info('SnapRobotServer is now running on: http://{}:{}\n'.format(snap_host, snap_port))
            logger.info('You can open Snap! interface with loaded blocks at "{}"\n'.format(url))

        if use_http:
            from pypot.server.httpserver import HTTPRobotServer
            poppy_creature.http = HTTPRobotServer(poppy_creature, http_host, http_port,
                                                  cross_domain_origin="*", quiet=http_quiet)
            logger.info('HTTPRobotServer is now running on: http://{}:{}\n'.format(http_host, http_port))

        if use_remote:
            from pypot.server import RemoteRobotServer
            poppy_creature.remote = RemoteRobotServer(poppy_creature, remote_host, remote_port)
            logger.info('RemoteRobotServer is now running on: http://{}:{}\n'.format(remote_host, remote_port))

        if use_ws:
            from pypot.server import WsRobotServer
            poppy_creature.ws = WsRobotServer(poppy_creature, ws_host, ws_port)
            logger.info('Ws server is now running on: ws://{}:{}\n'.format(ws_host, ws_port))

        cls.setup(poppy_creature)

        if start_background_services:
            cls.start_background_services(poppy_creature)

        return poppy_creature

    @classmethod
    def start_background_services(cls, robot, services=['snap', 'http', 'remote', 'ws']):
        for service in services:
            if hasattr(robot, service):
                s = Thread(target=getattr(robot, service).run,
                           name='{}_server'.format(service))
                s.daemon = True
                s.start()
                logger.info("Starting {} service".format(service))

    @classmethod
    def setup(cls, robot):
        """ Classmethod used to specify your poppy creature.

        This is where you should attach any specific primitives for instance.

        """
        pass

    @classproperty
    @classmethod
    def default_config(cls):
        creature = camelcase_to_underscore(cls.__name__)
        base_path = os.path.dirname(__import__(creature).__file__)

        default_config = os.path.join(os.path.join(base_path, 'configuration'),
                                      '{}.json'.format(creature))

        with open(default_config) as f:
            return json.load(f)
